# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import List

import cv2
import numpy as np

from datumaro.components.annotation import Bbox, HashKey


def gen_hash_key(features: np.ndarray) -> HashKey:
    features = np.sign(features)
    hash_key = np.clip(features, 0, None)
    hash_key = hash_key.astype(np.uint8)
    hash_key = np.packbits(hash_key, axis=-1)
    return HashKey(hash_key)


def create_bboxes_with_rescaling(
    bboxes: np.ndarray, labels: np.ndarray, r_scale: float
) -> List[Bbox]:
    idx = 0
    anns = []
    for bbox, label in zip(bboxes, labels):
        points = r_scale * bbox[:4]
        x1, y1, x2, y2 = points
        conf = bbox[4]
        anns.append(
            Bbox(
                x=x1,
                y=y1,
                w=x2 - x1,
                h=y2 - y1,
                id=idx,
                label=label,
                attributes={"score": conf},
            )
        )
        idx += 1
    return anns


@dataclass
class RescaledImage:
    """Dataclass for a rescaled image.

    This dataclass represents a rescaled image along with the scaling information.

    Attributes:
        image: The rescaled image as a NumPy array.
        scale: The scale factor by which the image was resized to fit the model input size.
               The scale factor is the same for both height and width.

    Note:
        The `image` attribute stores the rescaled image as a NumPy array.
        The `scale` attribute represents the scale factor used to resize the image.
        The scale factor indicates how much the image was scaled to fit the model's input size.
    """

    image: np.ndarray
    scale: float


def rescale_img_keeping_aspect_ratio(
    img: np.ndarray, h_model: int, w_model: int, padding: bool = True
) -> RescaledImage:
    """
    Rescale image while maintaining its aspect ratio.

    This function rescales the input image to fit the requirements of the model input.
    It also attempts to preserve the original aspect ratio of the input image.
    If the aspect ratio of the input image does not match the aspect ratio required by the model,
    the function applies zero padding to the image boundaries to maintain the aspect ratio if `padding` option is true.

    Parameters:
        img: The image to be rescaled.
        h_model: The desired height of the image required by the model.
        w_model: The desired width of the image required by the model.
        padding: If true, pad the output image boundaries to make the output image size `(h_model, w_model).
            Otherwise, there is no pad, so that the output image size can be different with `(h_model, w_model)`.
    """
    assert len(img.shape) == 3

    h_img, w_img = img.shape[:2]

    scale = min(h_model / h_img, w_model / w_img)

    h_resize = min(int(scale * h_img), h_model)
    w_resize = min(int(scale * w_img), w_model)

    num_channel = img.shape[-1]

    if padding:
        resized_inputs = np.zeros((h_model, w_model, num_channel), dtype=np.uint8)

        resized_inputs[:h_resize, :w_resize, :] = cv2.resize(
            img,
            (w_resize, h_resize),
            interpolation=cv2.INTER_LINEAR,
        )
    else:
        resized_inputs = cv2.resize(
            img,
            (w_resize, h_resize),
            interpolation=cv2.INTER_LINEAR,
        )

    return RescaledImage(image=resized_inputs, scale=scale)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import shutil
import uuid

from datumaro.components.algorithms.hash_key_inference.explorer import Explorer
from datumaro.components.algorithms.hash_key_inference.hashkey_util import (
    check_and_convert_to_list,
    match_query_subset,
)
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.media import Image
from datumaro.util import str_to_bool
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Explore similar data of query in dataset",
        description="""
        Applies data exploration to a dataset for image/text query.
        The command can be useful if you have to find similar data in dataset.
        |n
        The current project (-p/--project) is used as a context for plugins
        and models. It is used when there is a dataset path in target.
        When not specified, the current project's working tree is used.|n
        |n
        Examples:|n
        - Explore top50 similar images of image query in COCO dataset:|n
        |s|s%(prog)s --query-img-path path/to/image.jpg -topk 50|n
        - Explore top50 similar images of text query, elephant, in COCO dataset:|n
        |s|s%(prog)s --query-str elephant -topk 50|n
        - Explore top50 similar images of image query list in COCO dataset:|n
        |s|s%(prog)s --query-img-path path/to/image1.jpg/ path/to/image2.jpg/ path/to/image3.jpg/ -topk 50|n
        - Explore top50 similar images of text query list in COCO dataset:|n
        |s|s%(prog)s --query-str motorcycle/ bus/ train/ -topk 50
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("target", nargs="?", help="Target dataset")
    query_parser = parser.add_mutually_exclusive_group(required=True)
    query_parser.add_argument(
        "--query-img-path",
        default=None,
        type=str,
        help="Image path of query to explore similar data",
    )
    query_parser.add_argument(
        "--query-item-id",
        default=None,
        type=str,
        help="Datasetitem id of query to explore similar data",
    )
    query_parser.add_argument(
        "--query-item-subset",
        default=None,
        type=str,
        help="Datasetitem subset of query to explore similar data",
    )
    query_parser.add_argument(
        "--query-str",
        default=None,
        type=str,
        help="Text to explore similar data",
    )
    parser.add_argument("-topk", type=int, dest="topk", help="Number of similar results")
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.add_argument(
        "-s",
        "--save",
        action="store_true",
        default=False,
        help="Save explorer result files on explore_result folder",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        default=None,
        help="Directory to save explore results " "(default: generate automatically)",
    )
    parser.add_argument(
        "--stage",
        type=str_to_bool,
        default=True,
        help="""
            Include this action as a project build step.
            If true, this operation will be saved in the project
            build tree, allowing to reproduce the resulting dataset later.
            Applicable only to main project targets (i.e. data sources
            and the 'project' target, but not intermediate stages)
            (default: %(default)s)
            """,
    )
    parser.set_defaults(command=explore_command)

    return parser


def get_sensitive_args():
    return {
        explore_command: [
            "target",
            "topk",
            "project_dir",
            "dst_dir",
        ]
    }


@scoped
def explore_command(args):
    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if args.project_dir:
            raise

    if args.target:
        targets = [args.target]
    else:
        targets = list(project.working_tree.sources)

    source_datasets = [parse_full_revpath(target, project)[0] for target in targets]

    explorer_args = {"save_hashkey": True}
    if project:
        build_tree = project.working_tree.clone()
        for target in targets:
            build_tree.build_targets.add_explore_stage(target, params=explorer_args)

    explorer = Explorer(*source_datasets)
    for dataset in source_datasets:
        dst_dir = dataset.data_path
        dataset.save(dst_dir, save_media=True, save_hashkey_meta=True)

    if args.stage and project:
        project.working_tree.config.update(build_tree.config)
        project.working_tree.save()

    if args.query_img_path:
        querys = check_and_convert_to_list(args.query_img_path)
        query_datasetitems = []
        for query_ in querys:
            query_datasetitem = DatasetItem(id=str(uuid.uuid4()), media=Image.from_file(query_))
            query_datasetitems.append(query_datasetitem)
    elif args.query_item_id:
        querys = (
            [args.query_item_id] if not isinstance(args.query_item_id, list) else args.query_item_id
        )
        query_datasetitems = []
        for query in querys:
            for dataset in source_datasets:
                query_datasetitem = match_query_subset(
                    query, dataset, subset=args.query_item_subset
                )
            query_datasetitems.append(query_datasetitem)
    elif args.query_str:
        query_datasetitems = args.query_str

    results = explorer.explore_topk(query_datasetitems, args.topk)

    result_path_list = []
    log.info(f"Most similar {args.topk} results of query in dataset")
    for result in results:
        path = getattr(result.media, "path", None)
        result_path_list.append(path)
        log.info(f"id: {result.id} | subset: {result.subset} | path : {path}")

    if args.save:
        saved_result_path = args.dst_dir or os.path.join(args.project_dir, "explore_result")
        if os.path.exists(saved_result_path):
            shutil.rmtree(saved_result_path)
        os.makedirs(saved_result_path)
        for result in results:
            saved_subset_path = os.path.join(saved_result_path, result.subset)
            if not os.path.exists(saved_subset_path):
                os.makedirs(saved_subset_path)
            shutil.copyfile(path, os.path.join(saved_subset_path, result.id + ".jpg"))

    return 0



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from inspect import isclass
from typing import (
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetBase, SubsetBase
from datumaro.components.exporter import Exporter
from datumaro.components.generator import DatasetGenerator
from datumaro.components.importer import Importer
from datumaro.components.launcher import Launcher
from datumaro.components.lazy_plugin import LazyPlugin
from datumaro.components.transformer import ItemTransform, Transform
from datumaro.components.validator import Validator

T = TypeVar("T")


class Registry(Generic[T]):
    def __init__(self):
        self._items: Dict[str, T] = {}

    def register(self, name: str, value: T) -> T:
        self._items[name] = value
        return value

    def unregister(self, name: str) -> Optional[T]:
        return self._items.pop(name, None)

    def get(self, key: str) -> T:
        """Returns a class or a factory function"""
        return self._items[key]

    def __getitem__(self, key: str) -> T:
        return self.get(key)

    def __contains__(self, key) -> bool:
        return key in self._items

    def __iter__(self) -> Iterator[str]:
        return iter(self._items)

    def items(self) -> Generator[Tuple[str, T], None, None]:
        for key in self:
            yield key, self.get(key)


class PluginRegistry(Registry[Type[CliPlugin]]):
    _ACCEPT: Type[CliPlugin] = None
    _SKIP: Optional[Iterable[Type[CliPlugin]]] = None
    _DECLINE: Optional[Type[CliPlugin]] = None

    def __init__(self):
        super().__init__()
        if self._ACCEPT is None:
            raise NotImplementedError(
                f"{self.__class__.__name__} requires an _ACCEPT class attribute"
                " to specify the accepted type of stored instances."
            )

    def _filter(self, t):
        skip = {self._SKIP} if isclass(self._SKIP) else set(self._SKIP or [])
        skip = tuple(skip | set((self._ACCEPT,)))
        if (
            not issubclass(t, self._ACCEPT)
            or t in skip
            or (self._DECLINE and issubclass(t, self._DECLINE))
        ):
            return False
        if getattr(t, "__not_plugin__", None):
            return False
        return True

    def get(self, key: str) -> Type[CliPlugin]:
        """Returns a class or a factory function"""
        item = self._items[key]
        if issubclass(item, LazyPlugin):
            return item.get_plugin_cls()
        return item

    def batch_register(self, values: Iterable[Type[CliPlugin]]):
        for v in values:
            if not self._filter(v):
                continue

            self.register(v.NAME, v)


class DatasetBaseRegistry(PluginRegistry):
    _ACCEPT = DatasetBase
    _SKIP = (SubsetBase, Transform, ItemTransform)
    _DECLINE = Transform


class ImporterRegistry(PluginRegistry):
    _ACCEPT = Importer

    def __init__(self):
        super().__init__()
        self.extension_groups = defaultdict(list)

    def register(
        self, name: str, value: Union[Type[Importer], LazyPlugin]
    ) -> Union[Type[Importer], LazyPlugin]:
        super().register(name, value)
        if issubclass(value, LazyPlugin):
            file_extensions = value.METADATA["file_extensions"]
        else:
            file_extensions = value.get_file_extensions()
        for extension in file_extensions:
            self.extension_groups[extension].append((name, value))
        return value


class LauncherRegistry(PluginRegistry):
    _ACCEPT = Launcher


class ExporterRegistry(PluginRegistry):
    _ACCEPT = Exporter


class GeneratorRegistry(PluginRegistry):
    _ACCEPT = DatasetGenerator


class TransformRegistry(PluginRegistry):
    _ACCEPT = Transform
    _SKIP = ItemTransform


class ValidatorRegistry(PluginRegistry):
    _ACCEPT = Validator



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.dataset import DEFAULT_FORMAT
from datumaro.components.environment import DEFAULT_ENVIRONMENT
from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.hl_ops import HLOps
from datumaro.components.merge.intersect_merge import IntersectMerge
from datumaro.components.project import ProjectBuildTargets
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter, join_cli_args
from ..util.errors import CliException
from ..util.project import generate_next_file_name, load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Merge few datasets or projects",
        description="""
        Merges multiple datasets into one and produces a new dataset.
        The command can be useful if you want to merge several datasets
        which have homogeneous or heterogeneous label categories.
        We provide three merge policies: 1) "exact" for homogenous datasets;
        2) "union", or 3) "intersect" for heterogenous datasets. The default merge
        policy is "union".
        |n
        |n
        In simple cases, when dataset images do not intersect and new
        labels are not added, we can use the "exact" merge policy for this situation.
        |n
        |n
        On the other hand, when two datasets have different label categories or
        have the same (id, subset) pairs in their dataset items. you have to use
        the "union" or "intersect" merge.
        |n
        The "union" merge is more simple way to merge them.
        It tries to obtain a union set from their label categories and attempts to merge
        their dataset items. However, if there exist the same (id, subset) pairs, it appends
        a suffix to each item's id to prevent the same (id, subset) pair in the merged dataset.
        |n
        The "intersect" policy is a more complicated way to merge heterogeneous datasets. If you want to merge
        the annotations in the same (id, subset) pairs which coming from the different datasets
        each other you wish to merge. You have to taking into consideration potential overlaps
        and conflicts between the annotations. "intersect" merge policy can try to find common
        ground by voting or return a list of conflicts.
        |n
        |n
        This command has multiple forms:|n
        1) %(prog)s <revpath>|n
        2) %(prog)s <revpath> <revpath> ...|n
        |n
        1 - Merges the current project's main target ('project')
        in the working tree with the specified dataset.|n
        2 - Merges the specified datasets.
        Note that the current project is not included in the list of merged
        sources automatically.|n
        |n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        The current project (-p/--project) is used as a context for plugins.
        It can be useful for dataset paths in targets. When not specified,
        the current project's working tree is used.|n
        |n
        The output format can be specified with the '-f/--format' option.
        Each dataset format has its own export
        options, which are passed after the '--' separator (see examples),
        pass '-- -h' for more info. If not stated otherwise, by default
        only annotations are exported; to include images pass
        '--save-media' parameter.|n
        |n
        Examples:|n
        - Merge annotations from 3 (or more) annotators:|n
        |s|s%(prog)s project1/ project2/ project3/|n
        |n
        - Merge datasets with varying merge policies:|n
        |s|s%(prog)s project1/ project2/ -m <union|intersect|exact>|n
        |n
        - Check groups of the merged dataset for consistency:|n
        |s|s|slook for groups consising of 'person', 'hand' 'head', 'foot'|n
        |s|s%(prog)s project1/ project2/ -g 'person,hand?,head,foot?'|n
        |n
        - Merge two datasets, specify formats:|n
        |s|s%(prog)s path/to/dataset1:voc path/to/dataset2:coco|n
        |n
        - Merge the current working tree and a dataset:|n
        |s|s%(prog)s path/to/dataset2:coco|n
        |n
        - Merge a source from a previous revision and a dataset:|n
        |s|s%(prog)s HEAD~2:source-2 path/to/dataset2:yolo
        |n
        - Merge datasets and save in different format:|n
        |s|s%(prog)s -f voc dataset1/:yolo path2/:coco -- --save-media
        """,
        formatter_class=MultilineFormatter,
    )

    def _group(s):
        return s.split(",")

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument("targets", nargs="+", help="Target dataset revpaths (repeatable)")
    parser.add_argument(
        "-m",
        "--merge-policy",
        default="union",
        type=str,
        help="Policy for how to merge datasets (default: %(default)s)",
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        default=None,
        help="Output directory (default: generate a new one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument(
        "-f", "--format", default=DEFAULT_FORMAT, help="Output format (default: %(default)s)"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the 'current' project (default: current dir)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for exporter (pass '-- -h' for help). "
        "Must be specified after the main command arguments and after "
        "the '--' separator",
    )

    intersect_args = parser.add_argument_group(
        title="intersect policy arguments",
        description="These parameters are optional for the intersect policy only.",
    )
    intersect_args.add_argument(
        "-iou",
        "--iou-thresh",
        default=0.25,
        type=float,
        help="IoU match threshold for segments (default: %(default)s)",
    )
    intersect_args.add_argument(
        "-oconf",
        "--output-conf-thresh",
        default=0.0,
        type=float,
        help="Confidence threshold for output " "annotations (default: %(default)s)",
    )
    intersect_args.add_argument(
        "--quorum",
        default=0,
        type=int,
        help="Minimum count for a label and attribute voting "
        "results to be counted (default: %(default)s)",
    )
    intersect_args.add_argument(
        "-g",
        "--groups",
        action="append",
        type=_group,
        help="A comma-separated list of labels in "
        "annotation groups to check. '?' postfix can be added to a label to "
        "make it optional in the group (repeatable)",
    )

    parser.set_defaults(command=merge_command)

    return parser


def get_sensitive_args():
    return {
        merge_command: ["targets", "project_dir", "dst_dir", "groups"],
    }


@scoped
def merge_command(args):
    # Workaround. Required positionals consume positionals from the end
    args._positionals += join_cli_args(args, "targets", "extra_args")

    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
        if pos == 0:
            raise argparse.ArgumentError(None, message="Expected at least 1 target argument")
    else:
        pos = len(args._positionals)
    args.targets = args._positionals[:pos] or [ProjectBuildTargets.MAIN_TARGET]
    args.extra_args = args._positionals[pos + has_sep :]

    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name("merged")
    dst_dir = osp.abspath(dst_dir)

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help and len(args.targets) == 1 and args.project_dir:
            raise

    if project is not None:
        env = project.env
    else:
        env = DEFAULT_ENVIRONMENT

    try:
        exporter = env.exporters[args.format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % args.format)

    export_args = exporter.parse_cmdline(args.extra_args)

    source_datasets = []
    try:
        if len(args.targets) == 1:
            source_datasets.append(project.working_tree.make_dataset())

        for t in args.targets:
            target_dataset, target_project = parse_full_revpath(t, project)
            if target_project:
                scope_add(target_project)
            source_datasets.append(target_dataset)
    except Exception as e:
        raise CliException(str(e))

    report_path = osp.join(dst_dir, "merge_report.json")
    options = (
        {
            "conf": IntersectMerge.Conf(
                pairwise_dist=args.iou_thresh,
                groups=args.groups or [],
                output_conf_thresh=args.output_conf_thresh,
                quorum=args.quorum,
            )
        }
        if args.merge_policy == "intersect"
        else {}
    )
    merged_dataset = HLOps.merge(
        *source_datasets, merge_policy=args.merge_policy, report_path=report_path, **options
    )

    merged_dataset.export(save_dir=dst_dir, format=exporter, **export_args)

    log.info("Merge results have been saved to '%s'" % dst_dir)
    log.info("Report has been saved to '%s'" % report_path)

    return 0



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from enum import Enum, auto


class SamplingMethod(Enum):
    topk = auto()
    lowk = auto()
    randk = auto()
    mixk = auto()
    randtopk = auto()


class Algorithm(Enum):
    entropy = auto()


class InferenceResultAnalyzer:
    """
    Basic interface for IRA (Inference Result Analyzer)
    """

    def __init__(self, dataset, inference):
        self.data = dataset
        self.inference = inference
        self.sampling_method = SamplingMethod

    def get_sample(self, method: str, k: int):
        raise NotImplementedError()



# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import errno
import glob
import os.path as osp
from typing import List, Optional

import nibabel as nib
import numpy as np

from datumaro.components.annotation import AnnotationType, ExtractedMask, LabelCategories
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import ImportContext, Importer
from datumaro.components.media import MultiframeImage


class BratsPath:
    IMAGES_DIR = "images"
    LABELS = "labels"
    DATA_EXT = ".nii.gz"


class BratsBase(SubsetBase):
    def __init__(self, path: str, *, ctx: Optional[ImportContext] = None):
        if not osp.isdir(path):
            raise NotADirectoryError(errno.ENOTDIR, "Can't find dataset directory", path)

        self._subset_suffix = osp.basename(path)[len(BratsPath.IMAGES_DIR) :]
        subset = None
        if self._subset_suffix == "Tr":
            subset = "train"
        elif self._subset_suffix == "Ts":
            subset = "test"
        super().__init__(subset=subset, media_type=MultiframeImage, ctx=ctx)

        self._root_dir = osp.dirname(path)
        self._categories = self._load_categories()
        self._items = list(self._load_items(path).values())

    def _load_categories(self):
        label_cat = LabelCategories()

        labels_path = osp.join(self._root_dir, BratsPath.LABELS)
        if osp.isfile(labels_path):
            with open(labels_path, encoding="utf-8") as f:
                for line in f:
                    label_cat.add(line.strip())

        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}

        for image_path in glob.glob(osp.join(path, f"*{BratsPath.DATA_EXT}")):
            data = nib.load(image_path).get_fdata()

            item_id = osp.basename(image_path)[: -len(BratsPath.DATA_EXT)]

            images = [0] * data.shape[2]
            for i in range(data.shape[2]):
                images[i] = data[:, :, i]

            items[item_id] = DatasetItem(
                id=item_id, subset=self._subset, media=MultiframeImage(images, path=image_path)
            )

        masks_dir = osp.join(self._root_dir, BratsPath.LABELS + self._subset_suffix)
        for mask in glob.glob(osp.join(masks_dir, f"*{BratsPath.DATA_EXT}")):
            data = nib.load(mask).get_fdata()

            item_id = osp.basename(image_path)[: -len(BratsPath.DATA_EXT)]

            if item_id not in items:
                items[item_id] = DatasetItem(id=item_id)

            anno = []
            for i in range(data.shape[2]):
                np_mask = data[:, :, i]
                classes = np.unique(np_mask)
                for class_id in classes:
                    anno.append(
                        ExtractedMask(
                            index_mask=np_mask,
                            index=class_id,
                            label=class_id,
                            attributes={"image_id": i},
                        )
                    )
                    self._ann_types.add(AnnotationType.mask)

            items[item_id].annotations = anno

        return items


class BratsImporter(Importer):
    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        with context.require_any():
            with context.alternative():
                context.require_file(f"*/*{BratsPath.DATA_EXT}")

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(path, "", "brats", filename=f"{BratsPath.IMAGES_DIR}*")

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [osp.splitext(BratsPath.DATA_EXT)[-1]]



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse

from datumaro.util.scope import scope_add, scoped

from ....util import MultilineFormatter
from ....util.project import load_project


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Create a revision",
        description="""
        Creates a new revision from the current state of the working directory.|n
        |n
        Examples:|n
        - Create a revision:|n
        |s|s%(prog)s -m "Added data"
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("-m", "--message", required=True, help="Commit message")
    parser.add_argument(
        "--allow-empty",
        action="store_true",
        help="Allow commits with no changes (default: %(default)s)",
    )
    parser.add_argument(
        "--allow-foreign",
        action="store_true",
        help="Allow commits with non-Datumaro changes (default: %(default)s)",
    )
    parser.add_argument(
        "--no-cache",
        action="store_true",
        help="Don't put committed datasets into cache, "
        "save only metainfo (default: %(default)s)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=commit_command)

    return parser


def get_sensitive_args():
    return {
        commit_command: [
            "message",
            "project_dir",
        ],
    }


@scoped
def commit_command(args):
    project = scope_add(load_project(args.project_dir))

    old_tree = project.head

    new_commit = project.commit(
        args.message,
        allow_empty=args.allow_empty,
        allow_foreign=args.allow_foreign,
        no_cache=args.no_cache,
    )

    new_tree = project.working_tree
    diff = project.diff(old_tree, new_tree)

    print("Moved to commit '%s' %s" % (new_commit, args.message))
    print(" %s targets changed" % len(diff))
    for t, s in diff.items():
        print(" %s %s" % (s.name, t))

    return 0



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import os

from datumaro.cli.util import MultilineFormatter


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="List supported import/export data formats",
        description="""
        List supported import/export data format names.
        For more detailed guides on each data format,
        please visit our documentation website:
        https://openvinotoolkit.github.io/datumaro/stable/docs/data-formats/formats
        |n
        |n
        Examples:|n
        - List supported import/export data format names:|n
        |s|s%(prog)s|n
        |n
        - List only supported import data format names:|n
        |s|s%(prog)s --list-import|n
        |n
        - List only supported export data format names:|n
        |s|s%(prog)s --list-export|n
        |n
        - List with comma delimiter:|n
        |s|s%(prog)s --delimiter ','
        """,
        formatter_class=MultilineFormatter,
    )

    group = parser.add_argument_group()
    exclusive_group = group.add_mutually_exclusive_group(required=False)

    exclusive_group.add_argument(
        "-li",
        "--list-import",
        action="store_true",
        help="List all supported import data format names",
    )
    exclusive_group.add_argument(
        "-le",
        "--list-export",
        action="store_true",
        help="List all supported export data format names",
    )
    parser.add_argument(
        "-d",
        "--delimiter",
        type=str,
        default=os.linesep,
        help="Seperator used to list data format names (default: \\n)",
    )

    parser.set_defaults(command=format_command)
    return parser


def get_sensitive_args():
    return {
        format_command: ["list", "show"],
    }


def format_command(args: argparse.Namespace) -> None:
    from datumaro.components.environment import DEFAULT_ENVIRONMENT

    delimiter = args.delimiter

    if args.list_import:
        builtin_readers = sorted(
            set(DEFAULT_ENVIRONMENT.importers) | set(DEFAULT_ENVIRONMENT.extractors)
        )
        print(delimiter.join(builtin_readers))
        return

    if args.list_export:
        builtin_writers = sorted(DEFAULT_ENVIRONMENT.exporters)
        print(delimiter.join(builtin_writers))
        return

    builtin_readers = sorted(
        set(DEFAULT_ENVIRONMENT.importers) | set(DEFAULT_ENVIRONMENT.extractors)
    )
    builtin_writers = sorted(DEFAULT_ENVIRONMENT.exporters)
    print(f"Supported import formats:\n{delimiter.join(builtin_readers)}")
    print(f"Supported export formats:\n{delimiter.join(builtin_writers)}")



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from datumaro.errors import DatasetImportError

_SIGNATURE = "signature:datumaro_binary"


class DatumaroBinaryPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"
    PCD_DIR = "point_clouds"
    VIDEO_DIR = "videos"
    MASKS_DIR = "masks"

    ANNOTATION_EXT = ".datum"
    IMAGE_EXT = ".jpg"
    MASK_EXT = ".png"
    SIGNATURE = _SIGNATURE
    SIGNATURE_LEN = len(_SIGNATURE)

    SECRET_KEY_FILE = "secret_key.txt"

    MAX_BLOB_SIZE = 2**20  # 1 Mega bytes
    MP_TIMEOUT = 300.0  # 5 minutes

    @classmethod
    def check_signature(cls, signature: str):
        if signature != cls.SIGNATURE:
            raise DatasetImportError(
                f"Input signature={signature} is not aligned with the ground truth signature={cls.SIGNATURE}"
            )



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import List, Union

import numpy as np
from ovmsclient import make_grpc_client, make_http_client
from ovmsclient.tfs_compat.grpc.serving_client import GrpcClient
from ovmsclient.tfs_compat.http.serving_client import HttpClient

from datumaro.components.abstracts.model_interpreter import LauncherInputType, ModelPred
from datumaro.components.errors import DatumaroError
from datumaro.plugins.inference_server_plugin.base import (
    LauncherForDedicatedInferenceServer,
    ProtocolType,
)

__all__ = ["OVMSLauncher"]

TClient = Union[GrpcClient, HttpClient]


class OVMSLauncher(LauncherForDedicatedInferenceServer[TClient]):
    """Inference launcher for OVMS (OpenVINO™ Model Server) (https://github.com/openvinotoolkit/model_server)

    Parameters:
        model_name: Name of the model. It should match with the model name loaded in the server instance.
        model_interpreter_path: Python source code path which implements a model interpreter.
            The model interpreter implement pre-processing of the model input and post-processing of the model output.
        model_version: Version of the model loaded in the server instance
        host: Host address of the server instance
        port: Port number of the server instance
        timeout: Timeout limit during communication between the client and the server instance
        tls_config: Configuration required if the server instance is in the secure mode
        protocol_type: Communication protocol type with the server instance
    """

    def _init_client(self) -> TClient:
        tls_config = self.tls_config.as_dict() if self.tls_config is not None else None

        if self.protocol_type == ProtocolType.grpc:
            return make_grpc_client(self.url, tls_config)
        if self.protocol_type == ProtocolType.http:
            return make_http_client(self.url, tls_config)

        raise NotImplementedError(self.protocol_type)

    def _check_server_health(self) -> None:
        status = self._client.get_model_status(
            model_name=self.model_name,
            model_version=self.model_version,
            timeout=self.timeout,
        )
        log.info(f"Health check succeeded: {status}")

    def _init_metadata(self):
        self._metadata = self._client.get_model_metadata(
            model_name=self.model_name,
            model_version=self.model_version,
            timeout=self.timeout,
        )
        log.info(f"Received metadata: {self._metadata}")

    def infer(self, inputs: LauncherInputType) -> List[ModelPred]:
        # Please see the following link for the input and output type of self._client.predict()
        # https://github.com/openvinotoolkit/model_server/blob/releases/2022/3/client/python/ovmsclient/lib/docs/grpc_client.md#method-predict
        # The input is Dict[str, np.ndarray].
        # The output is Dict[str, np.ndarray] (If the model has multiple outputs),
        # or np.ndarray (If the model has one single output).
        pred_inputs = {self._input_key: inputs} if isinstance(inputs, np.ndarray) else inputs
        results = self._client.predict(
            inputs=pred_inputs,
            model_name=self.model_name,
            model_version=self.model_version,
            timeout=self.timeout,
        )

        # If there is only one output key,
        # it returns `np.ndarray`` rather than `Dict[str, np.ndarray]`.
        # Please see ovmsclient.tfs_compat.grpc.responses.GrpcPredictResponse
        if isinstance(results, np.ndarray):
            results = {self._output_key: results}

        outputs_group_by_item = [
            {key: output for key, output in zip(results.keys(), outputs)}
            for outputs in zip(*results.values())
        ]

        return outputs_group_by_item

    @property
    def _input_key(self):
        if hasattr(self, "__input_key"):
            return self.__input_key

        metadata_inputs = self._metadata.get("inputs")

        if metadata_inputs is None:
            raise DatumaroError("Cannot get metadata of the outputs.")

        if len(metadata_inputs.keys()) > 1:
            raise DatumaroError(
                f"More than two model inputs are not allowed: {metadata_inputs.keys()}."
            )

        self.__input_key = next(iter(metadata_inputs.keys()))
        return self.__input_key

    @property
    def _output_key(self):
        if hasattr(self, "__output_key"):
            return self.__output_key

        metadata_outputs = self._metadata.get("outputs")

        if metadata_outputs is None:
            raise DatumaroError("Cannot get metadata of the outputs.")

        if len(metadata_outputs.keys()) > 1:
            raise DatumaroError(
                f"More than two model outputs are not allowed: {metadata_outputs.keys()}."
            )

        self.__output_key = next(iter(metadata_outputs.keys()))
        return self.__output_key



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import glob
import logging
import os
import os.path as osp
from typing import List, Optional, Type, TypeVar

from datumaro.components.annotation import AnnotationType, Bbox
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.errors import InvalidAnnotationError
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image
from datumaro.util.image import find_images

from .format import Kitti3DLabelMap, Kitti3dPath, make_kitti3d_categories

T = TypeVar("T")


class Kitti3dBase(SubsetBase):
    # https://www.cvlibs.net/datasets/kitti/eval_object.php?obj_benchmark=3d

    def __init__(
        self,
        path: str,
        *,
        subset: Optional[str] = None,
        ctx: Optional[ImportContext] = None,
    ):
        assert osp.isdir(path), path

        self._path = path

        if not subset:
            folder_path = path.rsplit(Kitti3dPath.LABEL_DIR, 1)[0]
            img_dir = osp.join(folder_path, Kitti3dPath.IMAGE_DIR)
            if any(os.path.isdir(os.path.join(img_dir, item)) for item in os.listdir(img_dir)):
                subset = osp.split(path)[-1]
                self._path = folder_path
        super().__init__(subset=subset, ctx=ctx)

        self._categories = make_kitti3d_categories(Kitti3DLabelMap)
        self._items = self._load_items()

    def _load_items(self) -> List[DatasetItem]:
        items = []

        image_dir = osp.join(self._path, Kitti3dPath.IMAGE_DIR)
        image_path_by_id = {
            osp.split(osp.splitext(osp.relpath(p, image_dir))[0])[-1]: p
            for p in find_images(image_dir, recursive=True)
        }

        if self._subset == "default":
            ann_dir = osp.join(self._path, Kitti3dPath.LABEL_DIR)
        else:
            ann_dir = osp.join(self._path, Kitti3dPath.LABEL_DIR, self._subset)

        label_categories = self._categories[AnnotationType.label]

        for labels_path in sorted(glob.glob(osp.join(ann_dir, "**", "*.txt"), recursive=True)):
            item_id = osp.splitext(osp.relpath(labels_path, ann_dir))[0]
            anns = []

            try:
                with open(labels_path, "r", encoding="utf-8") as f:
                    lines = f.readlines()
            except IOError as e:
                logging.error(f"Error reading file {labels_path}: {e}")
                continue

            for line_idx, line in enumerate(lines):
                line = line.split()
                if len(line) not in [15, 16]:
                    logging.warning(
                        f"Unexpected line length {len(line)} in file {labels_path} at line {line_idx + 1}"
                    )
                    continue

                label_name = line[0]
                label_id = label_categories.find(label_name)[0]
                if label_id is None:
                    label_id = label_categories.add(label_name)

                try:
                    x1 = self._parse_field(line[4], float, "bbox left-top x")
                    y1 = self._parse_field(line[5], float, "bbox left-top y")
                    x2 = self._parse_field(line[6], float, "bbox right-bottom x")
                    y2 = self._parse_field(line[7], float, "bbox right-bottom y")

                    attributes = {
                        "truncated": self._parse_field(line[1], float, "truncated"),
                        "occluded": self._parse_field(line[2], int, "occluded"),
                        "alpha": self._parse_field(line[3], float, "alpha"),
                        "dimensions": [
                            self._parse_field(line[8], float, "height (in meters)"),
                            self._parse_field(line[9], float, "width (in meters)"),
                            self._parse_field(line[10], float, "length (in meters)"),
                        ],
                        "location": [
                            self._parse_field(line[11], float, "x (in meters)"),
                            self._parse_field(line[12], float, "y (in meters)"),
                            self._parse_field(line[13], float, "z (in meters)"),
                        ],
                        "rotation_y": self._parse_field(line[14], float, "rotation_y"),
                    }
                except ValueError as e:
                    logging.error(f"Error parsing line {line_idx + 1} in file {labels_path}: {e}")
                    continue

                anns.append(
                    Bbox(
                        x=x1,
                        y=y1,
                        w=x2 - x1,
                        h=y2 - y1,
                        id=line_idx,
                        attributes=attributes,
                        label=label_id,
                    )
                )
                self._ann_types.add(AnnotationType.bbox)

            image = image_path_by_id.pop(item_id, None)
            if image:
                image = Image.from_file(path=image)

            if self._subset == "default":
                calib_path = osp.join(self._path, Kitti3dPath.CALIB_DIR, item_id + ".txt")
            else:
                calib_path = osp.join(
                    self._path, Kitti3dPath.CALIB_DIR, self._subset, item_id + ".txt"
                )
            items.append(
                DatasetItem(
                    id=item_id,
                    subset=self._subset,
                    media=image,
                    attributes={"calib_path": calib_path},
                    annotations=anns,
                )
            )

        return items

    def _parse_field(self, value: str, desired_type: Type[T], field_name: str) -> T:
        try:
            return desired_type(value)
        except Exception as e:
            raise InvalidAnnotationError(
                f"Can't parse {field_name} from '{value}'. Expected {desired_type}"
            ) from e



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import struct
from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Optional, Set, Type

import pyarrow as pa

from datumaro.components.annotation import AnnotationType, Categories
from datumaro.components.dataset_base import (
    CategoriesInfo,
    DatasetBase,
    DatasetInfo,
    DatasetItem,
    IDataset,
    SubsetBase,
)
from datumaro.components.importer import ImportContext
from datumaro.components.media import Image, MediaElement, MediaType
from datumaro.components.merge.extractor_merger import check_identicalness
from datumaro.plugins.data_formats.arrow.format import DatumaroArrow
from datumaro.plugins.data_formats.datumaro.base import JsonReader
from datumaro.plugins.data_formats.datumaro_binary.mapper.common import DictMapper
from datumaro.util.definitions import DEFAULT_SUBSET_NAME

from .mapper.dataset_item import DatasetItemMapper


class ArrowSubsetBase(SubsetBase):
    __not_plugin__ = True

    def __init__(
        self,
        lookup: Dict[str, DatasetItem],
        infos: Dict[str, Any],
        categories: Dict[AnnotationType, Categories],
        subset: str,
        media_type: Type[MediaElement] = Image,
        ann_types: Set[AnnotationType] = None,
    ):
        super().__init__(
            length=len(lookup), subset=subset, media_type=media_type, ann_types=ann_types, ctx=None
        )

        self._lookup = lookup
        self._infos = infos
        self._categories = categories

    def __iter__(self) -> Iterator[DatasetItem]:
        for item in self._lookup.values():
            yield item

    def __len__(self) -> int:
        return len(self._lookup)

    def get(self, item_id: str, subset: Optional[str] = None) -> Optional[DatasetItem]:
        if subset != self._subset:
            return None

        try:
            return self._lookup[item_id]
        except KeyError:
            return None


@dataclass(frozen=True)
class Metadata:
    infos: Dict
    categories: Dict
    media_type: Type[MediaElement]


class ArrowBase(DatasetBase):
    def __init__(
        self,
        root_path: str,
        *,
        file_paths: List[str],
        ctx: Optional[ImportContext] = None,
    ):
        self._root_path = root_path
        tables = [pa.ipc.open_file(pa.memory_map(path, "r")).read_all() for path in file_paths]
        metadatas = [self._load_schema_metadata(table) for table in tables]

        table = pa.concat_tables(tables)
        subsets = table.column(DatumaroArrow.SUBSET_FIELD).unique().to_pylist()
        media_type = check_identicalness([metadata.media_type for metadata in metadatas])

        super().__init__(length=len(table), subsets=subsets, media_type=media_type, ctx=ctx)

        self._infos = check_identicalness([metadata.infos for metadata in metadatas])
        self._categories = check_identicalness([metadata.categories for metadata in metadatas])

        self._init_cache(file_paths, subsets)

    @staticmethod
    def _load_schema_metadata(table: pa.Table) -> Metadata:
        schema = table.schema

        _infos, _ = DictMapper.backward(schema.metadata.get(b"infos", b"\x00\x00\x00\x00"))
        infos = JsonReader._load_infos({"infos": _infos})

        _categories, _ = DictMapper.backward(
            schema.metadata.get(b"categories", b"\x00\x00\x00\x00")
        )
        categories = JsonReader._load_categories({"categories": _categories})

        (media_type,) = struct.unpack("<I", schema.metadata.get(b"media_type", b"\x00\x00\x00\x00"))
        media_type = MediaType(media_type).media

        return Metadata(infos=infos, categories=categories, media_type=media_type)

    def infos(self) -> DatasetInfo:
        return self._infos

    def categories(self) -> CategoriesInfo:
        return self._categories

    def __iter__(self) -> Iterator[DatasetItem]:
        for lookup in self._lookup.values():
            for item in lookup.values():
                yield item

    def _init_cache(self, file_paths: List[str], subsets: List[str]):
        self._lookup: Dict[str, Dict[str, DatasetItem]] = {subset: {} for subset in subsets}

        total = len(self)
        cnt = 0
        pbar = self._ctx.progress_reporter
        pbar.start(total=total, desc="Importing")

        ann_types = set()
        for table_path in file_paths:
            with pa.OSFile(table_path, "r") as source:
                with pa.ipc.open_file(source) as reader:
                    table = reader.read_all()
                    for idx in range(len(table)):
                        item = DatasetItemMapper.backward(idx, table, table_path)
                        self._lookup[item.subset][item.id] = item
                        for ann in item.annotations:
                            ann_types.add(ann.type)
                        pbar.report_status(cnt)
                        cnt += 1
        self._ann_types = ann_types

        self._subsets = {
            subset: ArrowSubsetBase(
                lookup=lookup,
                infos=self._infos,
                categories=self._categories,
                subset=self._subsets,
                media_type=self._media_type,
                ann_types=self._ann_types,
            )
            for subset, lookup in self._lookup.items()
        }

        pbar.finish()

    def get(self, item_id: str, subset: Optional[str] = None) -> Optional[DatasetItem]:
        subset = subset or DEFAULT_SUBSET_NAME

        try:
            return self._lookup[subset][item_id]
        except KeyError:
            return None

    @property
    def lookup(self) -> Dict[str, Dict[str, int]]:
        return self._lookup

    def subsets(self) -> Dict[str, IDataset]:
        return self._subsets

    def get_subset(self, name: str) -> IDataset:
        return self._subsets[name]



# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import hashlib
import logging as log
import warnings
from collections import defaultdict
from copy import deepcopy
from typing import Callable, Dict, Optional, Set, Tuple

import cv2
import numpy as np

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset_base import CategoriesInfo, DatasetItem, IDataset
from datumaro.components.errors import DatumaroError
from datumaro.components.media import Image
from datumaro.util.image import IMAGE_BACKEND, ImageColorChannel, decode_image_context


def mean_std(dataset: IDataset):
    counter = _MeanStdCounter()

    for item in dataset:
        counter.accumulate(item)

    return counter.get_result()


class _MeanStdCounter:
    """
    Computes unbiased mean and std. dev. for dataset images, channel-wise.
    """

    def __init__(self):
        self._stats = {}  # (id, subset) -> (pixel count, mean vec, std vec)

    def accumulate(self, item: DatasetItem):
        size = item.media.size
        if size is None:
            log.warning(
                "Item %s: can't detect image size, "
                "the image will be skipped from pixel statistics",
                item.id,
            )
            return
        count = np.prod(item.media.size)

        image = item.media.data
        if len(image.shape) == 2:
            image = image[:, :, np.newaxis]
        else:
            image = image[:, :, :3]
        # opencv is much faster than numpy here
        mean, std = cv2.meanStdDev(image.astype(np.double) / 255)

        self._stats[(item.id, item.subset)] = (count, mean, std)

    def get_result(self) -> Tuple[Tuple[float, float, float], Tuple[float, float, float]]:
        n = len(self._stats)

        if n == 0:
            return [0, 0, 0], [0, 0, 0]

        counts = np.empty(n, dtype=np.uint32)
        stats = np.empty((n, 2, 3), dtype=np.double)

        for i, v in enumerate(self._stats.values()):
            counts[i] = v[0]
            stats[i][0] = v[1].reshape(-1)
            stats[i][1] = v[2].reshape(-1)

        mean = lambda i, s: s[i][0]
        var = lambda i, s: s[i][1]

        # make variance unbiased
        np.multiply(np.square(stats[:, 1]), (counts / (counts - 1))[:, np.newaxis], out=stats[:, 1])

        # Use an online algorithm to:
        # - handle different image sizes
        # - avoid cancellation problem
        _, mean, var = self._compute_stats(stats, counts, mean, var)
        return mean * 255, np.sqrt(var) * 255

    # Implements online parallel computation of sample variance
    # https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    @staticmethod
    def _pairwise_stats(count_a, mean_a, var_a, count_b, mean_b, var_b):
        """
        Computes vector mean and variance.

        Needed do avoid catastrophic cancellation in floating point computations

        Returns:
            A tuple (total count, mean, variance)
        """

        # allow long arithmetics
        count_a = int(count_a)
        count_b = int(count_b)

        delta = mean_b - mean_a
        m_a = var_a * (count_a - 1)
        m_b = var_b * (count_b - 1)
        M2 = m_a + m_b + delta**2 * (count_a * count_b / (count_a + count_b))

        return (count_a + count_b, mean_a * 0.5 + mean_b * 0.5, M2 / (count_a + count_b - 1))

    @staticmethod
    def _compute_stats(stats, counts, mean_accessor, variance_accessor):
        """
        Recursively computes total count, mean and variance,
        does O(log(N)) calls.

        Args:
            stats: (float array of shape N, 2 * d, d = dimensions of values)
            count: (integer array of shape N)
            mean_accessor: (function(idx, stats)) to retrieve element mean
            variance_accessor: (function(idx, stats)) to retrieve element variance

        Returns:
            A tuple (total count, mean, variance)
        """

        m = mean_accessor
        v = variance_accessor
        n = len(stats)
        if n == 1:
            return counts[0], m(0, stats), v(0, stats)
        if n == 2:
            return __class__._pairwise_stats(
                counts[0], m(0, stats), v(0, stats), counts[1], m(1, stats), v(1, stats)
            )
        h = n // 2
        return __class__._pairwise_stats(
            *__class__._compute_stats(stats[:h], counts[:h], m, v),
            *__class__._compute_stats(stats[h:], counts[h:], m, v),
        )

    def __len__(self) -> int:
        return len(self._stats)


IMAGE_STATS_SCHEMA = {
    "dataset": {
        "images count": 0,
        "unique images count": 0,
        "repeated images count": 0,
        "repeated images": [],  # [[id1, id2], [id3, id4, id5], ...]
    },
    "subsets": {},
}


def compute_image_statistics(dataset: IDataset):
    if dataset.media_type() != Image:
        raise DatumaroError(
            f"Your dataset's media_type is {dataset.media_type()}, "
            "but only Image media_type is allowed."
        )

    stats = deepcopy(IMAGE_STATS_SCHEMA)

    stats_counter = _MeanStdCounter()
    unique_counter = _ItemMatcher()

    # NOTE: Force image color channel to RGB
    with decode_image_context(
        image_backend=IMAGE_BACKEND.get(),
        image_color_channel=ImageColorChannel.COLOR_RGB,
    ):
        for item in dataset:
            if not isinstance(item.media, Image):
                warnings.warn(
                    f"item (id: {item.id}, subset: {item.subset})"
                    f" has media_type, {item.media} but only Image media_type is allowed."
                )
                continue

            stats_counter.accumulate(item)
            unique_counter.process_item(item)

    def _extractor_stats(subset_name):
        sub_counter = _MeanStdCounter()
        sub_counter._stats = {
            k: v
            for k, v in stats_counter._stats.items()
            if subset_name and k[1] == subset_name or not subset_name
        }

        available = len(sub_counter._stats) != 0

        stats = {
            "images count": len(sub_counter),
        }

        if available:
            mean, std = sub_counter.get_result()

            stats.update(
                {
                    "image mean (RGB)": [float(v) for v in mean],
                    "image std (RGB)": [float(v) for v in std],
                }
            )
        else:
            stats.update(
                {
                    "image mean (RGB)": "n/a",
                    "image std (RGB)": "n/a",
                }
            )
        return stats

    for subset_name in dataset.subsets():
        stats["subsets"][subset_name] = _extractor_stats(subset_name)

    unique_items = unique_counter.get_result()
    repeated_items = [sorted(g) for g in unique_items.values() if 1 < len(g)]

    stats["dataset"].update(
        {
            "images count": len(stats_counter),
            "unique images count": len(unique_items),
            "repeated images count": len(repeated_items),
            "repeated images": repeated_items,  # [[id1, id2], [id3, id4, id5], ...]
        }
    )

    return stats


def compute_ann_statistics(dataset: IDataset):
    warnings.warn(
        "We are planning to change the type of stats['annotations']['labels']['distribution'] "
        "and stats['annotations']['segments']['pixel distribution'] from `list` to `(named) tuple`. "
        "If you are checking the types in your code, please revisit it after upgrading datumaro>=2.0.0.",
        FutureWarning,
    )
    labels: LabelCategories = dataset.categories().get(AnnotationType.label, LabelCategories())

    def get_label(ann):
        try:
            return labels.items[ann.label].name if ann.label is not None else None
        except IndexError:
            log.warning(f"annotation({ann}) has undefined label({ann.label})")
            return ann.label

    stats = {
        "images count": 0,
        "annotations count": 0,
        "unannotated images count": 0,
        "unannotated images": [],
        "annotations by type": {
            t.name: {
                "count": 0,
            }
            for t in AnnotationType
        },
        "annotations": {},
    }
    by_type = stats["annotations by type"]

    attr_template = {
        "count": 0,
        "values count": 0,
        "values present": set(),
        "distribution": {},  # value -> (count, total%)
    }
    label_stat = {
        "count": 0,
        "distribution": defaultdict(lambda: [0, 0]),  # label -> (count, total%)
        "attributes": {},
    }

    stats["annotations"]["labels"] = label_stat
    segm_stat = {
        "avg. area": 0,
        "area distribution": [],  # a histogram with 10 bins
        # (min, min+10%), ..., (min+90%, max) -> (count, total%)
        "pixel distribution": defaultdict(lambda: [0, 0]),  # label -> (count, total%)
    }
    stats["annotations"]["segments"] = segm_stat
    segm_areas = []
    pixel_dist = segm_stat["pixel distribution"]
    total_pixels = 0

    for l in labels.items:
        label_stat["distribution"][l.name] = [0, 0]
        pixel_dist[l.name] = [0, 0]

    for item in dataset:
        if len(item.annotations) == 0:
            stats["unannotated images"].append(item.id)
            continue

        for ann in item.annotations:
            by_type[ann.type.name]["count"] += 1

            if not hasattr(ann, "label") or ann.label is None:
                continue

            if ann.type in {AnnotationType.mask, AnnotationType.polygon, AnnotationType.bbox}:
                area = ann.get_area()
                segm_areas.append(area)
                pixel_dist[get_label(ann)][0] += int(area)

            label_stat["count"] += 1
            label_stat["distribution"][get_label(ann)][0] += 1

            for name, value in ann.attributes.items():
                if name.lower() in {"occluded", "visibility", "score", "id", "track_id"}:
                    continue
                attrs_stat = label_stat["attributes"].setdefault(name, deepcopy(attr_template))
                attrs_stat["count"] += 1
                attrs_stat["values present"].add(str(value))
                attrs_stat["distribution"].setdefault(str(value), [0, 0])[0] += 1

    stats["images count"] = len(dataset)

    stats["annotations count"] = sum(t["count"] for t in stats["annotations by type"].values())
    stats["unannotated images count"] = len(stats["unannotated images"])

    for label_info in label_stat["distribution"].values():
        label_info[1] = label_info[0] / (label_stat["count"] or 1)

    for label_attr in label_stat["attributes"].values():
        label_attr["values count"] = len(label_attr["values present"])
        label_attr["values present"] = sorted(label_attr["values present"])
        for attr_info in label_attr["distribution"].values():
            attr_info[1] = attr_info[0] / (label_attr["count"] or 1)

    # numpy.sum might be faster, but could overflow with large datasets.
    # Python's int can transparently mutate to be of indefinite precision (long)
    total_pixels = sum(int(a) for a in segm_areas)

    segm_stat["avg. area"] = total_pixels / (len(segm_areas) or 1.0)

    for label_info in segm_stat["pixel distribution"].values():
        label_info[1] = label_info[0] / (total_pixels or 1)

    if len(segm_areas) != 0:
        hist, bins = np.histogram(segm_areas)
        segm_stat["area distribution"] = [
            {
                "min": float(bin_min),
                "max": float(bin_max),
                "count": int(c),
                "percent": int(c) / len(segm_areas),
            }
            for c, (bin_min, bin_max) in zip(hist, zip(bins[:-1], bins[1:]))
        ]

    return stats


class _ItemMatcher:
    @staticmethod
    def _default_item_hash(item: DatasetItem):
        if not item.media or not item.media.has_data:
            if item.media and hasattr(item.media, "path"):
                return hash(item.media.path)

            log.warning(
                "Item (%s, %s) has no image " "info, counted as unique", item.id, item.subset
            )
            return None

        # Disable B303:md5, because the hash is not used in a security context
        return hashlib.md5(item.media.data.tobytes()).hexdigest()  # nosec

    def __init__(self, item_hash: Optional[Callable] = None):
        self._hash = item_hash or self._default_item_hash

        # hash -> [(id, subset), ...]
        self._unique: Dict[str, Set[Tuple[str, str]]] = {}

    def process_item(self, item: DatasetItem):
        h = self._hash(item)
        if h is None:
            h = str(id(item))  # anything unique

        self._unique.setdefault(h, set()).add((item.id, item.subset))

    def get_result(self):
        return self._unique


def find_unique_images(dataset: IDataset, item_hash: Optional[Callable] = None):
    matcher = _ItemMatcher(item_hash=item_hash)
    for item in dataset:
        matcher.process_item(item)
    return matcher.get_result()


def match_classes(a: CategoriesInfo, b: CategoriesInfo):
    a_label_cat = a.get(AnnotationType.label, LabelCategories())
    b_label_cat = b.get(AnnotationType.label, LabelCategories())

    a_labels = set(c.name for c in a_label_cat)
    b_labels = set(c.name for c in b_label_cat)

    matches = a_labels & b_labels
    a_unmatched = a_labels - b_labels
    b_unmatched = b_labels - a_labels
    return matches, a_unmatched, b_unmatched


def match_items_by_id(a: IDataset, b: IDataset):
    a_items = set((item.id, item.subset) for item in a)
    b_items = set((item.id, item.subset) for item in b)

    matches = a_items & b_items
    matches = [([m], [m]) for m in matches]
    a_unmatched = a_items - b_items
    b_unmatched = b_items - a_items
    return matches, a_unmatched, b_unmatched


def match_items_by_image_hash(a: IDataset, b: IDataset):
    a_hash = find_unique_images(a)
    b_hash = find_unique_images(b)

    a_items = set(a_hash)
    b_items = set(b_hash)

    matches = a_items & b_items
    a_unmatched = a_items - b_items
    b_unmatched = b_items - a_items

    matches = [(a_hash[h], b_hash[h]) for h in matches]
    a_unmatched = set(i for h in a_unmatched for i in a_hash[h])
    b_unmatched = set(i for h in b_unmatched for i in b_hash[h])

    return matches, a_unmatched, b_unmatched



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
from typing import Tuple

DEFAULT_SUBSET_NAME = "default"
BboxIntCoords = Tuple[int, int, int, int]  # (x, y, w, h)
SUBSET_NAME_BLACKLIST = {"labels", "images", "annotations", "instances"}
SUBSET_NAME_WHITELIST = {"train", "test", "val"}


def get_datumaro_cache_dir(
    _CACHE_DIR: str = osp.expanduser(os.getenv("XDG_CACHE_HOME", osp.join("~", ".cache")))
) -> str:
    """Get DATUMARO_CACHE_DIR. If it does not exists, create it."""
    DATUMARO_CACHE_DIR = osp.join(_CACHE_DIR, "datumaro")

    try:
        if not osp.exists(DATUMARO_CACHE_DIR):
            os.makedirs(DATUMARO_CACHE_DIR)
    except Exception as e:
        log.error(f"Cannot create DATUMARO_CACHE_DIR={DATUMARO_CACHE_DIR} since {e}.")

    return DATUMARO_CACHE_DIR



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

from typing import IO, Union

import yaml

from datumaro.components.errors import ImmutableObjectError


class Schema:
    class Item:
        def __init__(self, ctor, internal=False):
            self.ctor = ctor
            self.internal = internal

        def __call__(self, *args, **kwargs):
            return self.ctor(*args, **kwargs)

    def __init__(self, items=None, fallback=None):
        self._items = {}
        if items is not None:
            self._items.update(items)
        self._fallback = fallback

    def _get_items(self, allow_fallback=True):
        all_items = {}

        if allow_fallback and self._fallback is not None:
            all_items.update(self._fallback)
        all_items.update(self._items)

        return all_items

    def items(self, allow_fallback=True):
        return self._get_items(allow_fallback=allow_fallback).items()

    def keys(self, allow_fallback=True):
        return self._get_items(allow_fallback=allow_fallback).keys()

    def values(self, allow_fallback=True):
        return self._get_items(allow_fallback=allow_fallback).values()

    def __contains__(self, key):
        return key in self.keys()

    def __len__(self):
        return len(self._get_items())

    def __iter__(self):
        return iter(self._get_items())

    def __getitem__(self, key):
        default = object()
        value = self.get(key, default=default)
        if value is default:
            raise KeyError('Key "%s" does not exist' % (key,))
        return value

    def get(self, key, default=None):
        found = self._items.get(key, default)
        if found is not default:
            return found

        if self._fallback is not None:
            return self._fallback.get(key, default)
        return found


class SchemaBuilder:
    def __init__(self):
        self._items = {}

    def add(self, name, ctor=str, internal=False):
        if name in self._items:
            raise KeyError('Key "%s" already exists' % (name,))

        self._items[name] = Schema.Item(ctor, internal=internal)
        return self

    def build(self):
        return Schema(self._items)


class Config:
    def __init__(self, config=None, fallback=None, schema=None, mutable=True):
        # schema should be established first
        self.__dict__["_schema"] = schema
        self.__dict__["_mutable"] = True

        self.__dict__["_config"] = {}
        if fallback is not None:
            for k, v in fallback.items(allow_fallback=False):
                self.set(k, v)
        if config is not None:
            self.update(config)

        self.__dict__["_mutable"] = mutable

    def _items(self, allow_fallback=True, allow_internal=True):
        all_config = {}
        if allow_fallback and self._schema is not None:
            for key, item in self._schema.items():
                all_config[key] = item()
        all_config.update(self._config)

        if not allow_internal and self._schema is not None:
            for key, item in self._schema.items():
                if item.internal:
                    all_config.pop(key, None)
        return all_config

    def items(self, allow_fallback=True, allow_internal=True):
        return self._items(allow_fallback=allow_fallback, allow_internal=allow_internal).items()

    def keys(self, allow_fallback=True, allow_internal=True):
        return self._items(allow_fallback=allow_fallback, allow_internal=allow_internal).keys()

    def values(self, allow_fallback=True, allow_internal=True):
        return self._items(allow_fallback=allow_fallback, allow_internal=allow_internal).values()

    def __contains__(self, key):
        return key in self.keys()

    def __len__(self):
        return len(self.items())

    def __iter__(self):
        return iter(self.keys())

    def __getitem__(self, key):
        default = object()
        value = self.get(key, default=default)
        if value is default:
            raise KeyError('Key "%s" does not exist' % (key,))
        return value

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __getattr__(self, key):
        default = object()
        found = self.get(key, default=default)
        if found is default:
            raise AttributeError(key)
        return found

    def __setattr__(self, key, value):
        return self.set(key, value)

    def __eq__(self, other):
        try:
            keys = set(self.keys()) | set(other.keys())
            for k in keys:
                my_v = self[k]
                other_v = other[k]
                if my_v != other_v:
                    return False
            return True
        except (KeyError, AttributeError):
            return False

    def __repr__(self):
        return repr(dict(self))

    def update(self, other):
        for k, v in other.items():
            self.set(k, v)

    def remove(self, key):
        if not self._mutable:
            raise ImmutableObjectError()

        self._config.pop(key, None)

    def get(self, key, default=None):
        found = self._config.get(key, default)
        if found is not default:
            return found

        if self._schema is not None:
            found = self._schema.get(key, default)
            if found is not default:
                # ignore mutability
                found = found()
                self._config[key] = found
                return found

        return found

    def set(self, key, value):
        if not self._mutable:
            raise ImmutableObjectError()

        if self._schema is not None:
            if key not in self._schema:
                raise KeyError("Can not set key '%s' - schema mismatch: " "unknown key" % (key,))

            schema_entry = self._schema[key]
            schema_entry_instance = schema_entry()

            if isinstance(value, (dict, Config)) and isinstance(schema_entry_instance, Config):
                schema_entry_instance.update(value)
                value = schema_entry_instance
            elif not isinstance(value, type(schema_entry_instance)):
                raise ValueError(
                    "Can not set key '%s' - schema mismatch:"
                    "unexpected value type %s, expected %s"
                    % (key, type(value), type(schema_entry_instance))
                )

        self._config[key] = value
        return value

    @classmethod
    def parse(cls, path: Union[str, IO], *args, **kwargs):
        if isinstance(path, str):
            with open(path, "r", encoding="utf-8") as f:
                return cls(yaml.safe_load(f), *args, **kwargs)
        else:
            return cls(yaml.safe_load(path), *args, **kwargs)

    @staticmethod
    def yaml_representer(dumper, value):
        return dumper.represent_data(value._items(allow_internal=False, allow_fallback=False))

    def dump(self, path: Union[str, IO]):
        if isinstance(path, str):
            with open(path, "w", encoding="utf-8") as f:
                yaml.safe_dump(self, f)
        else:
            yaml.safe_dump(self, path)


yaml.add_multi_representer(Config, Config.yaml_representer, Dumper=yaml.SafeDumper)
yaml.add_multi_representer(
    tuple, lambda dumper, value: dumper.represent_data(list(value)), Dumper=yaml.SafeDumper
)


class DictConfig(Config):
    def __init__(self, default=None):
        super().__init__()
        self.__dict__["_default"] = default

    def set(self, key, value):
        if self._default is not None:
            schema_entry_instance = self._default(value)

            if isinstance(value, (dict, Config)) and isinstance(schema_entry_instance, Config):
                value = schema_entry_instance
            elif not isinstance(value, type(schema_entry_instance)):
                raise ValueError(
                    "Can not set key '%s' - schema mismatch:"
                    "unexpected value type %s, expected %s"
                    % (key, type(value), type(schema_entry_instance))
                )

        return super().set(key, value)



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

from .base import Merger
from .exact_merge import ExactMerge
from .intersect_merge import IntersectMerge
from .union_merge import UnionMerge

DEFAULT_MERGE_POLICY = "exact"


def get_merger(merge_policy: str = DEFAULT_MERGE_POLICY, *args, **kwargs) -> Merger:
    """
    Get :class:`Merger` according to `merge_policy`. You have to choose an appropriate `Merger`
    for your purpose. The available merge policies are "union", "intersect", and "exact".

    1. :class:`UnionMerge`

    Merge several datasets with "union" policy:

    - Label categories are merged according to the union of their label names.
    For example, if Dataset-A has {"car", "cat", "dog"} and Dataset-B has
    {"car", "bus", "truck"} labels, the merged dataset will have
    {"bust", "car", "cat", "dog", "truck"} labels.

    - If there are two or more dataset items whose (id, subset) pairs match each other,
    both are included in the merged dataset. At this time, since the same (id, subset)
    pair cannot be duplicated in the dataset, we add a suffix to the id of each source item.
    For example, if Dataset-A has DatasetItem(id="magic", subset="train") and Dataset-B has
    also DatasetItem(id="magic", subset="train"), the merged dataset will have
    DatasetItem(id="magic-0", subset="train") and DatasetItem(id="magic-1", subset="train").

    2. :class:`IntersectMerge`

    Merge several datasets with "intersect" policy:

    - If there are two or more dataset items whose (id, subset) pairs match each other,
    we can consider this as having an intersection in our dataset. This method merges
    the annotations of the corresponding :class:`DatasetItem` into one :class:`DatasetItem`
    to handle this intersection. The rule to handle merging annotations is provided by
    :class:`AnnotationMerger` according to their annotation types. For example,
    DatasetItem(id="item_1", subset="train", annotations=[Bbox(0, 0, 1, 1)]) from Dataset-A and
    DatasetItem(id="item_1", subset="train", annotations=[Bbox(.5, .5, 1, 1)]) from Dataset-B can be
    merged into DatasetItem(id="item_1", subset="train", annotations=[Bbox(0, 0, 1, 1)]).

    - Label categories are merged according to the union of their label names
    (Same as `UnionMerge`). For example, if Dataset-A has {"car", "cat", "dog"}
    and Dataset-B has {"car", "bus", "truck"} labels, the merged dataset will have
    {"bust", "car", "cat", "dog", "truck"} labels.

    - This merge has configuration parameters (`conf`) to control the annotation merge behaviors.

    For example,

    ```python
    merge = IntersectMerge(
        conf=IntersectMerge.Conf(
            pairwise_dist=0.25,
            groups=[],
            output_conf_thresh=0.0,
            quorum=0,
        )
    )
    ```

    For more details for the parameters, please refer to :class:`IntersectMerge.Conf`.

    3. :class:`ExactMerge`

    Merges several datasets using the "simple" algorithm:

    - All datasets should have the same categories
    - items are matched by (id, subset) pairs
    - matching items share the media info available:
        - nothing + nothing = nothing
        - nothing + something = something
        - something A + something B = conflict
    - annotations are matched by value and shared
    - in case of conflicts, throws an error
    """
    if merge_policy == "union":
        return UnionMerge(*args, **kwargs)
    elif merge_policy == "intersect":
        return IntersectMerge(*args, **kwargs)
    elif merge_policy == "exact":
        return ExactMerge(*args, **kwargs)

    raise ValueError(f"{merge_policy} is invalid Merger name.")



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
from typing import List, Optional, Tuple

from datumaro.components.dataset import Dataset

log.basicConfig(level=log.INFO)


try:
    import torch

    class BboxOverlaps2D:
        """2D Overlaps (e.g. IoUs, GIoUs) Calculator."""

        def __init__(self, scale=1.0, dtype=None):
            self.scale = scale
            self.dtype = dtype

        def _bbox_overlaps(self, bboxes1, bboxes2, eps=1e-6):
            """Calculate overlap between two set of bboxes."""
            # Either the boxes are empty or the length of boxes' last dimension is 4
            assert bboxes1.size(-1) == 4 or bboxes1.size(0) == 0
            assert bboxes2.size(-1) == 4 or bboxes2.size(0) == 0

            # Batch dim must be the same
            # Batch dim: (B1, B2, ... Bn)
            assert bboxes1.shape[:-2] == bboxes2.shape[:-2]
            batch_shape = bboxes1.shape[:-2]

            rows = bboxes1.size(-2)
            cols = bboxes2.size(-2)

            if rows * cols == 0:
                return bboxes1.new(batch_shape + (rows, cols))

            area1 = (bboxes1[..., 2] - bboxes1[..., 0]) * (bboxes1[..., 3] - bboxes1[..., 1])
            area2 = (bboxes2[..., 2] - bboxes2[..., 0]) * (bboxes2[..., 3] - bboxes2[..., 1])

            lt = torch.max(
                bboxes1[..., :, None, :2], bboxes2[..., None, :, :2]
            )  # [B, rows, cols, 2]
            rb = torch.min(
                bboxes1[..., :, None, 2:], bboxes2[..., None, :, 2:]
            )  # [B, rows, cols, 2]

            wh = (rb - lt).clamp(min=0)
            overlap = wh[..., 0] * wh[..., 1]

            union = area1[..., None] + area2[..., None, :] - overlap

            eps = union.new_tensor([eps])
            union = torch.max(union, eps)
            ious = overlap / union

            return ious

        def __call__(self, bboxes1: torch.Tensor, bboxes2: torch.Tensor):
            """Calculate IoU between 2D bboxes.

            Args:
                bboxes1 (Tensor or :obj:`BaseBoxes`): bboxes have shape (m, 4)
                    in <x1, y1, x2, y2> format, or shape (m, 5) in <x1, y1, x2,
                    y2, score> format.
                bboxes2 (Tensor or :obj:`BaseBoxes`): bboxes have shape (m, 4)
                    in <x1, y1, x2, y2> format, shape (m, 5) in <x1, y1, x2, y2,
                    score> format, or be empty. If ``is_aligned `` is ``True``,
                    then m and n must be equal.
                mode (str): "iou" (intersection over union), "iof" (intersection
                    over foreground), or "giou" (generalized intersection over
                    union).
                is_aligned (bool, optional): If True, then m and n must be equal.
                    Default False.

            Returns:
                Tensor: shape (m, n) if ``is_aligned `` is False else shape (m,)
            """
            assert bboxes1.size(-1) in [0, 4, 5]
            assert bboxes2.size(-1) in [0, 4, 5]
            if bboxes2.size(-1) == 5:
                bboxes2 = bboxes2[..., :4]
            if bboxes1.size(-1) == 5:
                bboxes1 = bboxes1[..., :4]

            return self._bbox_overlaps(bboxes1, bboxes2)

    class DataAwareAnchorGenerator:
        def __init__(
            self,
            img_size: Tuple[int, int],
            strides: List[int],
            scales: List[List[float]],
            ratios: List[List[float]],
            pos_thr: float,
            neg_thr: float,
            device: Optional[str] = "cpu",
        ):
            """Data-aware anchor generator for optimizing appropriate anchor scales and ratios.
            In general, anchor generator gets img_size and strides, and its assigner gets positive
            and negative thresholds for solving matching problem in object detection tasks.

            Args:
                img_size (Tuple[int, int]): Image size of height and width.
                strides (List[int]): Strides of feature map from feature pyramid network.
                This implicitly indicates receptive field size and base size of anchor generator.
                scales (List[float]): Initial scales for data-aware optimization.
                ratios (List[float]): Initial ratios for data-aware optimization.
                pos_thr (float): Positive threshold for matching in the following assigner.
                neg_thr (float): Negative threshold for matching in the following assigner.
                device (str): Device for computing gradient. Please refer to `torch.device`
            """
            assert len(strides) == len(scales) or len(strides) == len(ratios)
            assert pos_thr >= neg_thr

            self.img_size = img_size
            self.strides = strides
            self.scales = scales
            self.ratios = ratios
            self.pos_thr = pos_thr
            self.neg_thr = neg_thr

            self.device = device if torch.cuda.is_available() and device == "cuda" else "cpu"
            self.shifts = [self.get_shifts(stride) for stride in self.strides]

            self.iou_calculator = BboxOverlaps2D()

        def get_shifts(self, stride: int) -> torch.Tensor:
            """Bounding box proposals from anchor generator is composed of shifts and base anchors,
            where shifts is generated in mesh-grid manner and base anchors is combinations of ratios and
            scales. This function is to create mesh-grid shifts in the original image space.

            Args:
                stride (int): Strides of feature map from feature pyramid network.

            Returns:
                Tensor: Shift point coordinates.
            """

            def _meshgrid(x: torch.Tensor, y: torch.Tensor):
                # use shape instead of len to keep tracing while exporting to onnx
                xx = x.repeat(y.shape[0])
                yy = y.view(-1, 1).repeat(1, x.shape[0]).view(-1)
                return xx, yy

            feat_h, feat_w = self.img_size[0] // stride, self.img_size[1] // stride
            shift_x = torch.arange(0, feat_w) * stride
            shift_y = torch.arange(0, feat_h) * stride

            shift_xx, shift_yy = _meshgrid(shift_x, shift_y)
            shifts = torch.stack([shift_xx, shift_yy, shift_xx, shift_yy], dim=-1).to(self.device)

            return shifts

        def get_anchors(
            self, base_size: int, shifts: torch.Tensor, scales: torch.Tensor, ratios: torch.Tensor
        ) -> torch.Tensor:
            """This function is to create base anchors, which combinates ratios and scales.

            Args:
                base_size (int): Strides of feature map from feature pyramid network.
                shifts (Tensor): Shift point coordinates in the original image space.
                scales (Tensor): Scales for creating base anchors.
                ratios (Tensor): Ratios for creating base anchors.

            Returns:
                Tensor: Set of anchor bounding box coordinates.
            """
            w, h = base_size, base_size
            x_center, y_center = 0.5 * w, 0.5 * h

            h_ratios = torch.sqrt(ratios)
            w_ratios = 1 / h_ratios

            ws = (w * w_ratios[:, None] * scales[None, :]).view(-1)
            hs = (h * h_ratios[:, None] * scales[None, :]).view(-1)

            base_anchors = [
                x_center - 0.5 * ws,
                y_center - 0.5 * hs,
                x_center + 0.5 * ws,
                y_center + 0.5 * hs,
            ]
            base_anchors = torch.stack(base_anchors, dim=-1).to(self.device)

            anchors = base_anchors[None, :, :] + shifts[:, None, :]
            anchors = anchors.view(-1, 4)

            return anchors

        def initialize(self, targets, scales, ratios):
            init_ratios = []
            init_scales = []
            for level, base_size in enumerate(self.strides):
                representatives_dict = {
                    1: [0.5],
                    2: [0.4, 0.6],
                    3: [0.25, 0.5, 0.75],
                    4: [0.25, 0.4, 0.6, 0.75],
                    5: [0.1, 0.25, 0.5, 0.75, 0.9],
                }
                representatives = representatives_dict.get(len(scales[level]), None)
                if not representatives:
                    log.error("less than 5 number of strides are available.")

                anchors = self.get_anchors(
                    base_size, self.shifts[level], scales[level], ratios[level]
                )

                overlaps = self.iou_calculator(targets, anchors)
                gt_max_overlaps, _ = overlaps.max(dim=1)

                scale_samples, ratio_samples = [], []
                for target in targets[gt_max_overlaps > 0.1]:
                    h = target[3] - target[1]
                    w = target[2] - target[0]
                    r = h / w
                    affine_h = base_size * torch.sqrt(r)
                    affine_w = base_size / torch.sqrt(r)
                    s = max(h / affine_h, w / affine_w)

                    scale_samples.append(s)
                    ratio_samples.append(r)

                if len(scale_samples) < len(representatives):
                    init_scales.append(scales[level])
                    init_ratios.append(ratios[level])
                else:
                    init_scales.append(
                        [
                            torch.kthvalue(
                                torch.Tensor(scale_samples), int(p * len(scale_samples))
                            ).values.item()
                            for p in representatives
                        ]
                    )
                    init_ratios.append(
                        [
                            torch.kthvalue(
                                torch.Tensor(ratio_samples), int(p * len(ratio_samples))
                            ).values.item()
                            for p in representatives
                        ]
                    )

            return torch.Tensor(init_scales), torch.Tensor(init_ratios)

        def get_loss(self, targets: torch.Tensor, scales: torch.Tensor, ratios: torch.Tensor):
            """This function is to create base anchors, which combinates ratios and scales.

            Args:
                targets (Tensor): Set of target bounding box coordinates.
                scales (Tensor): Scales for creating base anchors.
                ratios (Tensor): Ratios for creating base anchors.

            Returns:
                float: Cost.
                float: Coverage rate.
            """
            all_anchors = []
            for level, base_size in enumerate(self.strides):
                anchors = self.get_anchors(
                    base_size, self.shifts[level], scales[level], ratios[level]
                )
                all_anchors.append(anchors)
            all_anchors = torch.cat(all_anchors, dim=0)

            overlaps = self.iou_calculator(targets, all_anchors)

            max_overlaps, _ = overlaps.max(dim=0)
            log.info(
                f"[ANCHOR] total: {overlaps.shape[1]}, "
                f"pos: {sum(max_overlaps >= self.pos_thr)}, "
                f"neg: {sum(max_overlaps < self.neg_thr)}"
            )

            gt_max_overlaps, _ = overlaps.max(dim=1)
            log.info(
                f"[GT] total: {overlaps.shape[0]}, "
                f"pos: {sum(gt_max_overlaps >= self.pos_thr)}, "
                f"neg: {sum(gt_max_overlaps < self.neg_thr)}"
            )

            cost = (
                500 * gt_max_overlaps[(gt_max_overlaps >= self.pos_thr)].sum()
                + max_overlaps[(max_overlaps >= self.pos_thr)].sum()
                - max_overlaps[(max_overlaps < self.neg_thr)].sum()
            )
            log.info(f"Cost: {cost}")

            num_gts = overlaps.size(0)
            pos_gts = 0
            for i in range(num_gts):
                if gt_max_overlaps[i] >= self.pos_thr:
                    max_iou_inds = (overlaps[i, :] == gt_max_overlaps[i]).nonzero()
                    if max_iou_inds.numel() != 0:
                        overlaps[:, max_iou_inds[0]] = 0
                        pos_gts += 1
            coverage_rate = pos_gts / num_gts
            log.info(f"Coverage rate: {coverage_rate}")

            return cost, coverage_rate

        def optimize(
            self,
            dataset: Dataset,
            subset: Optional[str] = None,
            batch_size: Optional[int] = 1024,
            learning_rate: Optional[float] = 0.1,
            num_iters: Optional[int] = 100,
        ):
            """This function is to create base anchors, which combinates ratios and scales.

            Args:
                dataset (Dataset): Desired dataset to optimize anchor scales and ratios.
                batch_size (int): Minibatch size.
                learning_rate (float): Learning rate.
                num_iters (int): Number of iterations.

            Returns:
                List[List[float]]: Optimized scales.
                List[List[float]]: Optimized ratios.
            """
            targets = []
            for item in dataset:
                if subset and item.subset != subset:
                    continue
                height, width = item.media.data.shape[:2]
                scale_h, scale_w = height // self.img_size[0], width // self.img_size[1]

                for ann in item.annotations:
                    x, y, w, h = ann.get_bbox()
                    x /= scale_w
                    y /= scale_h
                    w /= scale_w
                    h /= scale_h
                    targets.append([x, y, x + w, y + h])
            targets = torch.Tensor(targets).to(self.device)

            scales, ratios = torch.Tensor(self.scales).to(self.device), torch.Tensor(
                self.ratios
            ).to(self.device)
            scales, ratios = self.initialize(targets, scales, ratios)

            scales = scales.detach().requires_grad_(True)
            ratios = ratios.detach().requires_grad_(True)
            optimizer = torch.optim.Adam([scales, ratios], lr=learning_rate)

            opt_iter = 0
            opt_cov_rate = 0
            opt_scales = scales
            opt_ratios = ratios
            for iter in range(num_iters):
                random_indices = torch.randperm(targets.size(0))[:batch_size]
                cost, cov_rate = self.get_loss(
                    targets=targets[random_indices],
                    scales=scales,
                    ratios=ratios,
                )
                log.info(f"iter: {iter - 1} / cost: {cost} / scales: {scales} / ratios: {ratios}")

                if cov_rate >= opt_cov_rate:
                    opt_iter = iter
                    opt_scales = scales
                    opt_ratios = ratios

                optimizer.zero_grad()
                (-cost).backward()
                optimizer.step()

                ratios.data = torch.clamp(ratios.data, min=0.125, max=8)
                scales.data = torch.clamp(scales.data, min=0.0625, max=16)

            log.info(f"optimized scale/ratio: {opt_scales}/{opt_ratios} @ iter {opt_iter}")
            return opt_scales.tolist(), opt_ratios.tolist()

except ImportError:

    class DataAwareAnchorGenerator:
        def __init__(self):
            raise ImportError("Torch package not found. Cannot optimize anchor generator.")



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from typing import Dict, List, Optional, Type

from datumaro.components.format_detection import FormatDetectionConfidence, FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.merge.extractor_merger import ExtractorMerger
from datumaro.rust_api import JsonSectionPageMapper

from .format import DatumaroPath


class DatumaroImporter(Importer):
    PATH_CLS = DatumaroPath

    @classmethod
    def detect(
        cls,
        context: FormatDetectionContext,
    ) -> Optional[FormatDetectionConfidence]:
        annot_file = context.require_file(
            osp.join(cls.PATH_CLS.ANNOTATIONS_DIR, "*" + cls.PATH_CLS.ANNOTATION_EXT)
        )

        with context.probe_text_file(
            annot_file,
            'must be a JSON object with "categories" ' 'and "items" keys',
        ):
            fpath = osp.join(context.root_path, annot_file)
            page_mapper = JsonSectionPageMapper(fpath)
            sections = page_mapper.sections()
            if not {"categories", "items"} <= sections.keys():
                raise Exception

    @classmethod
    def find_sources(cls, path) -> List[Dict]:
        return cls._find_sources_recursive(
            path,
            cls.PATH_CLS.ANNOTATION_EXT,
            cls.NAME,
            dirname=cls.PATH_CLS.ANNOTATIONS_DIR,
        )

    @classmethod
    def get_file_extensions(cls) -> List[str]:
        return [cls.PATH_CLS.ANNOTATION_EXT]

    @property
    def can_stream(self) -> bool:
        return True

    def get_extractor_merger(self) -> Type[ExtractorMerger]:
        return ExtractorMerger



# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT

# ruff: noqa: F405

from .annotation import *
from .common import DictMapper, FloatListMapper, IntListMapper, Mapper, StringMapper
from .dataset_item import *
from .media import *

__all__ = [
    # anns
    "AnnotationListMapper",
    "LabelMapper",
    "MaskMapper",
    "RleMaskMapper",
    "PointsMapper",
    "PolygonMapper",
    "PolyLineMapper",
    "BboxMapper",
    "CaptionMapper",
    "Cuboid3dMapper",
    "EllipseMapper",
    "Cuboid2DMapper",
    # common
    "Mapper",
    "DictMapper",
    "StringMapper",
    "IntListMapper",
    "FloatListMapper",
    # dataset_item
    "DatasetItemMapper",
    # media
    "MediaMapper",
]



# Copyright (C) 2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp

from datumaro.components.annotation import AnnotationType, LabelCategories


class Kitti3dPath:
    PCD_DIR = osp.join("velodyne")
    IMAGE_DIR = "image_2"
    LABEL_DIR = "label_2"
    CALIB_DIR = "calib"


Kitti3DLabelMap = [
    "DontCare",
    "Car",
    "Pedestrian",
    "Van",
    "Truck",
    "Cyclist",
    "Sitter",
    "Train",
    "Motorcycle",
    "Bus",
    "Misc",
]


def make_kitti3d_categories(label_map=None):
    if label_map is None:
        label_map = Kitti3DLabelMap

    categories = {}
    common_attrs = {"truncated", "occluded", "alpha", "dimensions", "location", "rotation_y"}
    label_categories = LabelCategories(attributes=common_attrs)
    for label in label_map:
        label_categories.add(label)
    categories[AnnotationType.label] = label_categories

    return categories



# Copyright (C) 2023-2024 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
from textwrap import wrap
from typing import Dict, List, Set, Tuple
from unittest import TestCase

from attr import attrib, attrs
from tabulate import tabulate

from datumaro.cli.util.project import generate_next_file_name
from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.annotations.matcher import LineMatcher, PointsMatcher, match_segments_pair
from datumaro.components.dataset import Dataset
from datumaro.components.operations import (
    compute_ann_statistics,
    compute_image_statistics,
    match_items_by_id,
    match_items_by_image_hash,
)
from datumaro.components.shift_analyzer import ShiftAnalyzer
from datumaro.util import dump_json_file, filter_dict, find
from datumaro.util.annotation_util import find_instances, max_bbox
from datumaro.util.attrs_util import default_if_none


@attrs
class DistanceComparator:
    iou_threshold = attrib(converter=float, default=0.5)

    def match_annotations(self, item_a, item_b):
        return {t: self._match_ann_type(t, item_a, item_b) for t in AnnotationType}

    def _match_ann_type(self, t, *args):
        # pylint: disable=no-value-for-parameter
        if t == AnnotationType.label:
            return self.match_labels(*args)
        elif t == AnnotationType.bbox:
            return self.match_boxes(*args)
        elif t == AnnotationType.polygon:
            return self.match_polygons(*args)
        elif t == AnnotationType.mask:
            return self.match_masks(*args)
        elif t == AnnotationType.points:
            return self.match_points(*args)
        elif t == AnnotationType.polyline:
            return self.match_lines(*args)
        # pylint: enable=no-value-for-parameter
        else:
            raise NotImplementedError("Unexpected annotation type %s" % t)

    @staticmethod
    def _get_ann_type(t, item):
        return [a for a in item.annotations if a.type == t]

    def match_labels(self, item_a, item_b):
        a_labels = set(a.label for a in self._get_ann_type(AnnotationType.label, item_a))
        b_labels = set(a.label for a in self._get_ann_type(AnnotationType.label, item_b))

        matches = a_labels & b_labels
        a_unmatched = a_labels - b_labels
        b_unmatched = b_labels - a_labels
        return matches, a_unmatched, b_unmatched

    def _match_segments(self, t, item_a, item_b):
        a_boxes = self._get_ann_type(t, item_a)
        b_boxes = self._get_ann_type(t, item_b)
        return match_segments_pair(a_boxes, b_boxes, dist_thresh=self.iou_threshold)

    def match_polygons(self, item_a, item_b):
        return self._match_segments(AnnotationType.polygon, item_a, item_b)

    def match_masks(self, item_a, item_b):
        return self._match_segments(AnnotationType.mask, item_a, item_b)

    def match_boxes(self, item_a, item_b):
        return self._match_segments(AnnotationType.bbox, item_a, item_b)

    def match_points(self, item_a, item_b):
        a_points = self._get_ann_type(AnnotationType.points, item_a)
        b_points = self._get_ann_type(AnnotationType.points, item_b)

        instance_map = {}
        for s in [item_a.annotations, item_b.annotations]:
            s_instances = find_instances(s)
            for inst in s_instances:
                inst_bbox = max_bbox(inst)
                for ann in inst:
                    instance_map[id(ann)] = [inst, inst_bbox]
        matcher = PointsMatcher(instance_map=instance_map)

        return match_segments_pair(
            a_points, b_points, dist_thresh=self.iou_threshold, distance=matcher.distance
        )

    def match_lines(self, item_a, item_b):
        a_lines = self._get_ann_type(AnnotationType.polyline, item_a)
        b_lines = self._get_ann_type(AnnotationType.polyline, item_b)

        matcher = LineMatcher()

        return match_segments_pair(
            a_lines, b_lines, dist_thresh=self.iou_threshold, distance=matcher.distance
        )


@attrs
class EqualityComparator:
    match_images: bool = attrib(kw_only=True, default=False)
    ignored_fields = attrib(kw_only=True, factory=set, validator=default_if_none(set))
    ignored_attrs = attrib(kw_only=True, factory=set, validator=default_if_none(set))
    ignored_item_attrs = attrib(kw_only=True, factory=set, validator=default_if_none(set))
    all = attrib(kw_only=True, default=False)

    _test: TestCase = attrib(init=False)
    errors: list = attrib(init=False)

    def __attrs_post_init__(self):
        self._test = TestCase()
        self._test.maxDiff = None

    def _match_items(self, a, b):
        if self.match_images:
            return match_items_by_image_hash(a, b)
        else:
            return match_items_by_id(a, b)

    def _compare_categories(self, a, b):
        test = self._test
        errors = self.errors

        try:
            test.assertEqual(sorted(a, key=lambda t: t.value), sorted(b, key=lambda t: t.value))
        except AssertionError as e:
            errors.append({"type": "categories", "message": str(e)})

        if AnnotationType.label in a:
            try:
                test.assertEqual(
                    a[AnnotationType.label].items,
                    b[AnnotationType.label].items,
                )
            except AssertionError as e:
                errors.append({"type": "labels", "message": str(e)})
        if AnnotationType.mask in a:
            try:
                test.assertEqual(
                    a[AnnotationType.mask].colormap,
                    b[AnnotationType.mask].colormap,
                )
            except AssertionError as e:
                errors.append({"type": "colormap", "message": str(e)})
        if AnnotationType.points in a:
            try:
                test.assertEqual(
                    a[AnnotationType.points].items,
                    b[AnnotationType.points].items,
                )
            except AssertionError as e:
                errors.append({"type": "points", "message": str(e)})

    def _compare_annotations(self, a, b):
        ignored_fields = self.ignored_fields
        ignored_attrs = self.ignored_attrs

        a_fields = {k: None for k in a.as_dict() if k in ignored_fields}
        b_fields = {k: None for k in b.as_dict() if k in ignored_fields}
        if "attributes" not in ignored_fields:
            a_fields["attributes"] = filter_dict(a.attributes, ignored_attrs)
            b_fields["attributes"] = filter_dict(b.attributes, ignored_attrs)

        result = a.wrap(**a_fields) == b.wrap(**b_fields)

        return result

    def _compare_items(self, item_a, item_b):
        test = self._test

        a_id = (item_a.id, item_a.subset)
        b_id = (item_b.id, item_b.subset)

        matched = []
        unmatched = []
        errors = []

        try:
            test.assertEqual(
                filter_dict(item_a.attributes, self.ignored_item_attrs),
                filter_dict(item_b.attributes, self.ignored_item_attrs),
            )
        except AssertionError as e:
            errors.append({"type": "item_attr", "a_item": a_id, "b_item": b_id, "message": str(e)})

        b_annotations = item_b.annotations[:]
        for ann_a in item_a.annotations:
            ann_b_candidates = [x for x in item_b.annotations if x.type == ann_a.type]

            ann_b = find(
                enumerate(self._compare_annotations(ann_a, x) for x in ann_b_candidates),
                lambda x: x[1],
            )
            if ann_b is None:
                unmatched.append(
                    {
                        "item": a_id,
                        "source": "a",
                        "ann": str(ann_a),
                    }
                )
                continue
            else:
                ann_b = ann_b_candidates[ann_b[0]]

            b_annotations.remove(ann_b)  # avoid repeats
            matched.append({"a_item": a_id, "b_item": b_id, "a": str(ann_a), "b": str(ann_b)})

        for ann_b in b_annotations:
            unmatched.append({"item": b_id, "source": "b", "ann": str(ann_b)})

        return matched, unmatched, errors

    @staticmethod
    def _print_output(output: dict):
        print("Found:")
        print("The first project has %s unmatched items" % len(output.get("a_extra_items", [])))
        print("The second project has %s unmatched items" % len(output.get("b_extra_items", [])))
        print("%s item conflicts" % len(output.get("errors", [])))
        print("%s matching annotations" % len(output.get("matches", [])))
        print("%s mismatching annotations" % len(output.get("mismatches", [])))

    def compare_datasets(self, a, b):
        self.errors = []
        errors = self.errors

        self._compare_categories(a.categories(), b.categories())

        matched = []
        unmatched = []

        matches, a_unmatched, b_unmatched = self._match_items(a, b)

        if a.categories().get(AnnotationType.label) != b.categories().get(AnnotationType.label):
            output = {
                "mismatches": unmatched,
                "a_extra_items": sorted(a_unmatched),
                "b_extra_items": sorted(b_unmatched),
                "errors": errors,
            }
            if self.all:
                output["matches"] = matched

            self._print_output(output)
            return output

        _dist = lambda s: len(s[1]) + len(s[2])
        for a_ids, b_ids in matches:
            # build distance matrix
            match_status = {}  # (a_id, b_id): [matched, unmatched, errors]
            a_matches = {a_id: None for a_id in a_ids}
            b_matches = {b_id: None for b_id in b_ids}

            for a_id in a_ids:
                item_a = a.get(*a_id)
                candidates = {}

                for b_id in b_ids:
                    item_b = b.get(*b_id)

                    i_m, i_um, i_err = self._compare_items(item_a, item_b)
                    candidates[b_id] = [i_m, i_um, i_err]

                    if len(i_um) == 0:
                        a_matches[a_id] = b_id
                        b_matches[b_id] = a_id
                        matched.extend(i_m)
                        errors.extend(i_err)
                        break

                match_status[a_id] = candidates

            # assign
            for a_id in a_ids:
                if len(b_ids) == 0:
                    break

                # find the closest, ignore already assigned
                matched_b = a_matches[a_id]
                if matched_b is not None:
                    continue
                min_dist = -1
                for b_id in b_ids:
                    if b_matches[b_id] is not None:
                        continue
                    d = _dist(match_status[a_id][b_id])
                    if d < min_dist and 0 <= min_dist:
                        continue
                    min_dist = d
                    matched_b = b_id

                if matched_b is None:
                    continue
                a_matches[a_id] = matched_b
                b_matches[matched_b] = a_id

                m = match_status[a_id][matched_b]
                matched.extend(m[0])
                unmatched.extend(m[1])
                errors.extend(m[2])

            a_unmatched |= set(a_id for a_id, m in a_matches.items() if not m)
            b_unmatched |= set(b_id for b_id, m in b_matches.items() if not m)

        output = {
            "mismatches": unmatched,
            "a_extra_items": sorted(a_unmatched),
            "b_extra_items": sorted(b_unmatched),
            "errors": errors,
        }
        if self.all:
            output["matches"] = matched
        self._print_output(output)
        return output

    @staticmethod
    def save_compare_report(
        output: Dict,
        report_dir: str,
    ) -> None:
        """Saves the comparison report to JSON and text files.

        Args:
            output: A dictionary containing the comparison data.
            report_dir: A string representing the directory to save the report files.
        """
        os.makedirs(report_dir, exist_ok=True)
        output_file = osp.join(
            report_dir,
            generate_next_file_name("equality_compare", ext=".json", basedir=report_dir),
        )

        log.info(f"Saving compare json to {output_file}")
        dump_json_file(output_file, output, indent=True)


@attrs
class TableComparator:
    """
    Class for comparing datasets and generating comparison report table.
    """

    @staticmethod
    def _extract_labels(dataset: Dataset) -> Set[str]:
        """Extracts labels from the dataset.

        Args:
            dataset: An instance of a Dataset class.

        Returns:
            A set of labels present in the dataset.
        """
        label_cat = dataset.categories().get(AnnotationType.label, LabelCategories())
        return set(c.name for c in label_cat)

    @staticmethod
    def _compute_statistics(dataset: Dataset) -> Tuple[Dict, Dict]:
        """Computes image and annotation statistics of the dataset.

        Args:
            dataset: An instance of a Dataset class.

        Returns:
            A tuple containing image statistics and annotation statistics.
        """
        image_stats = compute_image_statistics(dataset)
        ann_stats = compute_ann_statistics(dataset)
        return image_stats, ann_stats

    def _analyze_dataset(self, dataset: Dataset) -> Tuple[str, Set[str], Dict, Dict]:
        """Analyzes the dataset to get labels, format, and statistics.

        Args:
            dataset: An instance of a Dataset class.

        Returns:
            A tuple containing Dataset format, set of label names, image statistics,
            and annotation statistics.
        """
        dataset_format = dataset.format
        dataset_labels = self._extract_labels(dataset)
        image_stats, ann_stats = self._compute_statistics(dataset)
        return dataset_format, dataset_labels, image_stats, ann_stats

    @staticmethod
    def _create_table(headers: List[str], rows: List[List[str]]) -> str:
        """Creates a table with the given headers and rows using the tabulate module.

        Args:
            headers: A list containing table headers.
            rows: A list containing table rows.

        Returns:
            A string representation of the table.
        """

        def wrapfunc(item):
            """Wrap a item consisted of text, returning a list of wrapped lines."""
            max_len = 35
            return "\n".join(wrap(item, max_len))

        wrapped_rows = []
        for row in rows:
            new_row = [wrapfunc(item) for item in row]
            wrapped_rows.append(new_row)

        return tabulate(wrapped_rows, headers, tablefmt="grid")

    @staticmethod
    def _create_dict(rows: List[List[str]]) -> Dict[str, List[str]]:
        """Creates a dictionary from the rows of the table.

        Args:
            rows: A list containing table rows.

        Returns:
            A dictionary where the key is the first element of a row and the value is
            the rest of the row.
        """
        data_dict = {row[0]: row[1:] for row in rows[1:]}
        return data_dict

    def _create_high_level_comparison_table(
        self, first_info: Tuple, second_info: Tuple
    ) -> Tuple[str, Dict]:
        """Generates a high-level comparison table.

        Args:
            first_info: A tuple containing information about the first dataset.
            second_info: A tuple containing information about the second dataset.

        Returns:
            A tuple containing the table as a string and a dictionary representing the data
            of the table.
        """
        first_format, first_labels, first_image_stats, first_ann_stats = first_info
        second_format, second_labels, second_image_stats, second_ann_stats = second_info

        headers = ["Field", "First", "Second"]

        rows = [
            ["Format", first_format, second_format],
            ["Number of classes", str(len(first_labels)), str(len(second_labels))],
            [
                "Common classes",
                ", ".join(sorted(list(first_labels.intersection(second_labels)))),
                ", ".join(sorted(list(second_labels.intersection(first_labels)))),
            ],
            ["Classes", ", ".join(sorted(first_labels)), ", ".join(sorted(second_labels))],
            [
                "Images count",
                str(first_image_stats["dataset"]["images count"]),
                str(second_image_stats["dataset"]["images count"]),
            ],
            [
                "Unique images count",
                str(first_image_stats["dataset"]["unique images count"]),
                str(second_image_stats["dataset"]["unique images count"]),
            ],
            [
                "Repeated images count",
                str(first_image_stats["dataset"]["repeated images count"]),
                str(second_image_stats["dataset"]["repeated images count"]),
            ],
            [
                "Annotations count",
                str(first_ann_stats["annotations count"]),
                str(second_ann_stats["annotations count"]),
            ],
            [
                "Unannotated images count",
                str(first_ann_stats["unannotated images count"]),
                str(second_ann_stats["unannotated images count"]),
            ],
        ]

        table = self._create_table(headers, rows)
        data_dict = self._create_dict(rows)

        return table, data_dict

    def _create_mid_level_comparison_table(
        self, first_info: Tuple, second_info: Tuple
    ) -> Tuple[str, Dict]:
        """Generates a mid-level comparison table.

        Args:
            first_info: A tuple containing information about the first dataset.
            second_info: A tuple containing information about the second dataset.

        Returns:
            A tuple containing the table as a string and a dictionary representing the data
            of the table.
        """
        _, _, first_image_stats, first_ann_stats = first_info
        _, _, second_image_stats, second_ann_stats = second_info

        headers = ["Field", "First", "Second"]

        rows = []

        first_subsets = sorted(list(first_image_stats["subsets"].keys()))
        second_subsets = sorted(list(second_image_stats["subsets"].keys()))

        subset_names = first_subsets.copy()
        subset_names.extend(item for item in second_subsets if item not in first_subsets)

        for subset_name in subset_names:
            first_subset_data = first_image_stats["subsets"].get(subset_name, {})
            second_subset_data = second_image_stats["subsets"].get(subset_name, {})
            mean_str_first = (
                ", ".join(f"{val:6.2f}" for val in first_subset_data.get("image mean (RGB)", []))
                if "image mean (RGB)" in first_subset_data
                else ""
            )
            std_str_first = (
                ", ".join(f"{val:6.2f}" for val in first_subset_data.get("image std (RGB)", []))
                if "image std" in first_subset_data
                else ""
            )
            mean_str_second = (
                ", ".join(f"{val:6.2f}" for val in second_subset_data.get("image mean (RGB)", []))
                if "image mean (RGB)" in second_subset_data
                else ""
            )
            std_str_second = (
                ", ".join(f"{val:6.2f}" for val in second_subset_data.get("image std", []))
                if "image std (RGB)" in second_subset_data
                else ""
            )
            rows.append([f"{subset_name} - Image Mean (RGB)", mean_str_first, mean_str_second])
            rows.append([f"{subset_name} - Image Std (RGB)", std_str_first, std_str_second])

        first_labels = sorted(list(first_ann_stats["annotations"]["labels"]["distribution"].keys()))
        second_labels = sorted(
            list(second_ann_stats["annotations"]["labels"]["distribution"].keys())
        )

        label_names = first_labels.copy()
        label_names.extend(item for item in second_labels if item not in first_labels)

        for label_name in label_names:
            count_dist_first = first_ann_stats["annotations"]["labels"]["distribution"].get(
                label_name, [0, 0.0]
            )
            count_dist_second = second_ann_stats["annotations"]["labels"]["distribution"].get(
                label_name, [0, 0.0]
            )
            count_first, dist_first = count_dist_first if count_dist_first[0] != 0 else ["", ""]
            count_second, dist_second = count_dist_second if count_dist_second[0] != 0 else ["", ""]
            rows.append(
                [
                    f"Label - {label_name}",
                    f"imgs: {count_first}, percent: {dist_first:.4f}" if count_first != "" else "",
                    f"imgs: {count_second}, percent: {dist_second:.4f}"
                    if count_second != ""
                    else "",
                ]
            )

        table = self._create_table(headers, rows)
        data_dict = self._create_dict(rows)

        return table, data_dict

    def _create_low_level_comparison_table(
        self, first_dataset: Dataset, second_dataset: Dataset
    ) -> Tuple[str, Dict]:
        """Generates a low-level comparison table.

        Args:
            first_dataset: The first dataset to compare.
            second_dataset: The second dataset to compare.

        Returns:
            A tuple containing the table as a string and a dictionary representing the data
            of the table.
        """
        shift_analyzer = ShiftAnalyzer()
        cov_shift = shift_analyzer.compute_covariate_shift([first_dataset, second_dataset])
        label_shift = shift_analyzer.compute_label_shift([first_dataset, second_dataset])

        headers = ["Field", "Value"]

        rows = [
            ["Covariate shift", str(cov_shift)],
            ["Label shift", str(label_shift)],
        ]

        table = self._create_table(headers, rows)
        data_dict = self._create_dict(rows)

        return table, data_dict

    def compare_datasets(
        self, first: Dataset, second: Dataset, mode: str = "all"
    ) -> Tuple[str, str, str, Dict]:
        """Compares two datasets and generates comparison reports.

        Args:
            first: The first dataset to compare.
            second: The second dataset to compare.

        Returns:
            A tuple containing high-level table, mid-level table, low-level table, and a
            dictionary representation of the comparison.
        """
        first_info = self._analyze_dataset(first)
        second_info = self._analyze_dataset(second)

        high_level_table, high_level_dict = None, {}
        mid_level_table, mid_level_dict = None, {}
        low_level_table, low_level_dict = None, {}

        if mode in ["high", "all"]:
            high_level_table, high_level_dict = self._create_high_level_comparison_table(
                first_info, second_info
            )
        if mode in ["mid", "all"]:
            mid_level_table, mid_level_dict = self._create_mid_level_comparison_table(
                first_info, second_info
            )
        if mode in ["low", "all"]:
            low_level_table, low_level_dict = self._create_low_level_comparison_table(first, second)

        comparison_dict = dict(
            high_level=high_level_dict, mid_level=mid_level_dict, low_level=low_level_dict
        )

        print(f"High-level comparison:\n{high_level_table}\n")
        print(f"Mid-level comparison:\n{mid_level_table}\n")
        print(f"Low-level comparison:\n{low_level_table}\n")

        return high_level_table, mid_level_table, low_level_table, comparison_dict

    @staticmethod
    def save_compare_report(
        high_level_table: str,
        mid_level_table: str,
        low_level_table: str,
        comparison_dict: Dict,
        report_dir: str,
    ) -> None:
        """Saves the comparison report to JSON and text files.

        Args:
            high_level_table: High-level comparison table as a string.
            mid_level_table: Mid-level comparison table as a string.
            low_level_table: Low-level comparison table as a string.
            comparison_dict: A dictionary containing the comparison data.
            report_dir: A string representing the directory to save the report files.
        """
        os.makedirs(report_dir, exist_ok=True)
        json_output_file = osp.join(
            report_dir, generate_next_file_name("table_compare", ext=".json", basedir=report_dir)
        )
        txt_output_file = osp.join(
            report_dir, generate_next_file_name("table_compare", ext=".txt", basedir=report_dir)
        )

        log.info(f"Saving compare json to {json_output_file}")
        log.info(f"Saving compare table to {txt_output_file}")

        dump_json_file(json_output_file, comparison_dict, indent=True)
        with open(txt_output_file, "w") as f:
            f.write(f"High-level Comparison:\n{high_level_table}\n\n")
            f.write(f"Mid-level Comparison:\n{mid_level_table}\n\n")
            f.write(f"Low-level Comparison:\n{low_level_table}\n\n")



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import copy
import logging as log
from enum import Enum, auto
from math import gcd

import numpy as np

from datumaro.components.annotation import AnnotationType
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DEFAULT_SUBSET_NAME
from datumaro.components.transformer import Transform
from datumaro.util import cast

NEAR_ZERO = 1e-7


class SplitTask(Enum):
    classification = auto()
    detection = auto()
    segmentation = auto()
    reid = auto()


class Split(Transform, CliPlugin):
    """
    - classification split |n
    |s|s|s|sSplits dataset into subsets(train/val/test) in class-wise manner. |n
    |s|s|s|sSplits dataset images in the specified ratio, keeping the initial class |n
    |s|s|s|sdistribution.|n
    |n
    - detection & segmentation split |n
    |s|s|s|sEach image can have multiple object annotations - |n
    |s|s|s|s(bbox, mask, polygon). Since an image shouldn't be included |n
    |s|s|s|sin multiple subsets at the same time, and image annotations |n
    |s|s|s|sshouldn't be split, in general, dataset annotations are unlikely |n
    |s|s|s|sto be split exactly in the specified ratio. |n
    |s|s|s|sThis split tries to split dataset images as close as possible |n
    |s|s|s|sto the specified ratio, keeping the initial class distribution.|n
    |n
    - reidentification split |n
    |s|s|s|sIn this task, the test set should consist of images of unseen|n
    |s|s|s|speople or objects during the training phase.|n
    |s|s|s|sThis function splits a dataset in the following way:|n
    |n
    |s|s1. Splits the dataset into 'train + val' and 'test' sets |n
    |s|s|s|s|sbased on person or object ID.|n
    |s|s2. Splits 'test' set into 'test-gallery' and 'test-query' sets |n
    |s|s|s|s|sin class-wise manner.|n
    |s|s3. Splits the 'train + val' set into 'train' and 'val' sets |n
    |s|s|s|s|sin the same way.|n
    |n
    The final subsets would be|n
    'train', 'val', 'test-gallery' and 'test-query'. |n
    |n
    Notes:|n
    |s|s- Each image is expected to have only one Annotation. Unlabeled or |n
    |s|s|s|smulti-labeled images will be split into subsets randomly. |n
    |s|s- If Labels also have attributes, also splits by attribute values.|n
    |s|s- If there is not enough images in some class or attributes group, |n
    |s|s|s|sthe split ratio can't be guaranteed. |n
    |s|s|s|sIn reidentification task, |n
    |s|s- Object ID can be described by Label, or by attribute (--attr parameter)|n
    |s|s- The splits of the test set are controlled by '--query' parameter |n
    |s|s|s|sGallery ratio would be 1.0 - query.|n
    |n
    Example:|n

    .. code-block::

    |s|s%(prog)s -t classification --subset train:.5 --subset val:.2 --subset test:.3 |n
    |s|s%(prog)s -t detection --subset train:.5 --subset val:.2 --subset test:.3 |n
    |s|s%(prog)s -t segmentation --subset train:.5 --subset val:.2 --subset test:.3 |n
    |s|s%(prog)s -t reid --subset train:.5 --subset val:.2 --subset test:.3 --query .5 |n
    |n
    Example: use 'person_id' attribute for splitting|n

    .. code-block::

    |s|s%(prog)s --attr person_id
    """

    _default_split = [("train", 0.5), ("val", 0.2), ("test", 0.3)]
    _default_query_ratio = 0.5

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "-t",
            "--task",
            default=SplitTask.classification.name,
            choices=[t.name for t in SplitTask],
            help="(one of {}; default: %(default)s)".format(", ".join(t.name for t in SplitTask)),
        )
        parser.add_argument(
            "-s",
            "--subset",
            action="append",
            type=cls._split_arg,
            dest="splits",
            help="Subsets in the form: '<subset>:<ratio>' "
            "(repeatable, default: %s)" % dict(cls._default_split),
        )
        parser.add_argument(
            "--query",
            type=float,
            default=None,
            help="Query ratio in the test set (default: %.3f)" % cls._default_query_ratio,
        )
        parser.add_argument(
            "--attr",
            type=str,
            dest="attr_for_id",
            default=None,
            help="Attribute name representing the ID (default: use label)",
        )
        parser.add_argument("--seed", type=int, help="Random seed")
        return parser

    @staticmethod
    def _split_arg(s):
        parts = s.split(":")
        if len(parts) != 2:
            import argparse

            raise argparse.ArgumentTypeError()
        return (parts[0], float(parts[1]))

    def __init__(self, dataset, task, splits, query=None, attr_for_id=None, seed=None):
        super().__init__(dataset)

        if splits is None:
            splits = self._default_split

        self.task = task
        self.splitter = self._get_splitter(task, dataset, splits, seed, query, attr_for_id)
        self._initialized = False
        self._subsets = self.splitter._subsets

    @staticmethod
    def _get_splitter(task, dataset, splits, seed, query, attr_for_id):
        if task == SplitTask.classification.name:
            splitter = _ClassificationSplit(dataset=dataset, splits=splits, seed=seed)
        elif task in {SplitTask.detection.name, SplitTask.segmentation.name}:
            splitter = _InstanceSpecificSplit(dataset=dataset, splits=splits, seed=seed, task=task)
        elif task == SplitTask.reid.name:
            splitter = _ReidentificationSplit(
                dataset=dataset,
                splits=splits,
                seed=seed,
                query=query,
                attr_for_id=attr_for_id,
            )
        else:
            raise Exception(
                f"Unknown task '{task}', available "
                f"splitter format: {[a.name for a in SplitTask]}"
            )
        return splitter

    def __iter__(self):
        # lazy splitting
        if self._initialized is False:
            self.splitter._split_dataset()
            self._initialized = True
        for i, item in enumerate(self._extractor):
            yield self.wrap_item(item, subset=self.splitter._find_split(i))

    def get_subset(self, name):
        # lazy splitting
        if self._initialized is False:
            self.splitter._split_dataset()
            self._initialized = True
        return super().get_subset(name)

    def subsets(self):
        # lazy splitting
        if self._initialized is False:
            self.splitter._split_dataset()
            self._initialized = True
        return super().subsets()


class _TaskSpecificSplit:
    def __init__(self, dataset, splits, seed, restrict=False):
        self._extractor = dataset

        snames, sratio, subsets = self._validate_splits(splits, restrict)

        self._snames = snames
        self._sratio = sratio

        self._seed = seed

        # remove subset name restriction
        # https://github.com/openvinotoolkit/datumaro/issues/194
        self._subsets = subsets
        self._parts = []
        self._length = "parent"

        self._initialized = False

    def _set_parts(self, by_splits):
        self._parts = []
        for subset in self._subsets:
            self._parts.append((set(by_splits[subset]), subset))

    @staticmethod
    def _get_uniq_annotations(dataset):
        annotations = []
        unlabeled_or_multi = []

        for idx, item in enumerate(dataset):
            labels = [a for a in item.annotations if a.type == AnnotationType.label]
            if len(labels) == 1:
                annotations.append(labels[0])
            else:
                unlabeled_or_multi.append(idx)

        return annotations, unlabeled_or_multi

    @staticmethod
    def _validate_splits(splits, restrict=False):
        snames = []
        ratios = []
        subsets = set()
        valid = ["train", "val", "test"]
        for subset, ratio in splits:
            # remove subset name restriction
            # https://github.com/openvinotoolkit/datumaro/issues/194
            if restrict:
                assert subset in valid, "Subset name must be one of %s, got %s" % (
                    valid,
                    subset,
                )
            assert (
                0.0 <= ratio and ratio <= 1.0
            ), "Ratio is expected to be in the range " "[0, 1], but got %s for %s" % (
                ratio,
                subset,
            )
            # ignore near_zero ratio because it may produce partition error.
            if ratio > NEAR_ZERO:
                # handling duplication
                if subset in snames:
                    raise Exception("Subset (%s) is duplicated" % subset)
                snames.append(subset)
                ratios.append(float(ratio))
            subsets.add(subset)

        ratios = np.array(ratios)

        total_ratio = np.sum(ratios)
        if not abs(total_ratio - 1.0) <= NEAR_ZERO:
            raise Exception(
                "Sum of ratios is expected to be 1, got %s, which is %s" % (splits, total_ratio)
            )

        return snames, ratios, subsets

    @staticmethod
    def _get_required(ratio):
        if len(ratio) < 2:
            return 1

        for scale in [10, 100]:
            farray = np.array(ratio) * scale
            iarray = farray.astype(int)
            if np.array_equal(iarray, farray):
                break

        # find gcd
        common_divisor = iarray[0]
        for val in iarray[1:]:
            common_divisor = gcd(common_divisor, val)

        required = np.sum(np.array(iarray / common_divisor).astype(int))

        return required

    @staticmethod
    def _get_sections(dataset_size, ratio):
        n_splits = [int(np.around(dataset_size * r)) for r in ratio[:-1]]
        n_splits.append(dataset_size - np.sum(n_splits))

        # if there are splits with zero samples even if ratio is not 0,
        # borrow one from the split who has one or more.
        for ii, num_split in enumerate(n_splits):
            if num_split == 0 and NEAR_ZERO < ratio[ii]:
                midx = np.argmax(n_splits)
                if n_splits[midx] > 0:
                    n_splits[ii] += 1
                    n_splits[midx] -= 1
        sections = np.add.accumulate(n_splits[:-1])
        return sections, n_splits

    @staticmethod
    def _group_by_attr(items):
        """
        Args:
            items: list of (idx_img, ann). ann is the annotation from Label object.
        Returns:
            by_attributes: dict of { combination-of-attrs : list of index }
        """

        # float--> numerical, others(int, string, bool) --> categorical
        def _is_float(value):
            if isinstance(value, str):
                casted = cast(value, float)
                if casted is not None:
                    if cast(casted, str) == value:
                        return True
                return False
            elif isinstance(value, float):
                cast(value, float)
                return True
            return False

        # group by attributes
        by_attributes = dict()
        for idx_img, ann in items:
            # ignore numeric attributes
            filtered = {}
            for attr, value in ann.attributes.items():
                if _is_float(value):
                    continue
                filtered[attr] = value
            attributes = tuple(sorted(filtered.items()))
            if attributes not in by_attributes:
                by_attributes[attributes] = []
            by_attributes[attributes].append(idx_img)

        return by_attributes

    def _split_by_attr(self, datasets, snames, ratio, out_splits, merge_small_classes=True):
        def _split_indice(indice):
            sections, _ = self._get_sections(len(indice), ratio)
            splits = np.array_split(indice, sections)
            for subset, split in zip(snames, splits):
                if 0 < len(split):
                    out_splits[subset].extend(split)

        required = self._get_required(ratio)
        rest = []
        for _, items in datasets.items():
            np.random.shuffle(items)
            by_attributes = self._group_by_attr(items)
            attr_combinations = list(by_attributes.keys())
            np.random.shuffle(attr_combinations)  # add randomness
            for attr in attr_combinations:
                indice = by_attributes[attr]
                quo = len(indice) // required
                if quo > 0:
                    filtered_size = quo * required
                    _split_indice(indice[:filtered_size])
                    rest.extend(indice[filtered_size:])
                else:
                    rest.extend(indice)

                quo = len(rest) // required
                if quo > 0:
                    filtered_size = quo * required
                    _split_indice(rest[:filtered_size])
                    rest = rest[filtered_size:]

            if not merge_small_classes and len(rest) > 0:
                _split_indice(rest)
                rest = []

        if len(rest) > 0:
            _split_indice(rest)

    def _split_unlabeled(self, unlabeled, by_splits):
        """
        split unlabeled data into subsets (detection, classification)
        Args:
            unlabeled: list of index of unlabeled or multi-labeled data
            by_splits: splits up to now
        Returns:
            by_splits: final splits
        """
        dataset_size = len(self._extractor)
        _, n_splits = list(self._get_sections(dataset_size, self._sratio))
        counts = [len(by_splits[sname]) for sname in self._snames]
        expected = [max(0, v) for v in np.subtract(n_splits, counts)]
        sections = np.add.accumulate(expected[:-1])
        np.random.shuffle(unlabeled)
        splits = np.array_split(unlabeled, sections)
        for subset, split in zip(self._snames, splits):
            if 0 < len(split):
                by_splits[subset].extend(split)

    def _find_split(self, index):
        for subset_indices, subset in self._parts:
            if index in subset_indices:
                return subset
        return DEFAULT_SUBSET_NAME  # all the possible remainder --> default

    def _split_dataset(self):
        raise NotImplementedError()


class _ClassificationSplit(_TaskSpecificSplit):
    """
    Splits dataset into subsets(train/val/test) in class-wise manner. |n
    Splits dataset images in the specified ratio, keeping the initial class
    distribution.|n
    |n
    Notes:|n
    |s|s- Each image is expected to have only one Label. Unlabeled or
    |s|s|s|smulti-labeled images will be split into subsets randomly. |n
    |s|s- If Labels also have attributes, also splits by attribute values.|n
    |s|s- If there is not enough images in some class or attributes group,
    |s|s|s|sthe split ratio can't be guaranteed.|n
    |n
    Example:|n

    .. code-block::

    |s|s%(prog)s -t classification --subset train:.5 --subset val:.2 --subset test:.3
    """

    def __init__(self, dataset, splits, seed=None):
        """
        Parameters
        ----------
        dataset : Dataset
        splits : list
            A list of (subset(str), ratio(float))
            The sum of ratios is expected to be 1.
        seed : int
            optional
        """
        super().__init__(dataset, splits, seed)

    def _split_dataset(self):
        np.random.seed(self._seed)

        # support only single label for a DatasetItem
        # 1. group by label
        by_labels = dict()
        annotations, unlabeled = self._get_uniq_annotations(self._extractor)

        for idx, ann in enumerate(annotations):
            label = getattr(ann, "label", None)
            if label not in by_labels:
                by_labels[label] = []
            by_labels[label].append((idx, ann))

        by_splits = dict()
        for subset in self._subsets:
            by_splits[subset] = []

        # 2. group by attributes
        self._split_by_attr(by_labels, self._snames, self._sratio, by_splits)

        # 3. split unlabeled data
        if len(unlabeled) > 0:
            self._split_unlabeled(unlabeled, by_splits)

        # 4. set parts
        self._set_parts(by_splits)


class _ReidentificationSplit(_TaskSpecificSplit):
    """
    Splits a dataset for re-identification task.|n
    Produces a split with a specified ratio of images, avoiding having same
    labels in different subsets.|n
    |n
    In this task, the test set should consist of images of unseen
    people or objects during the training phase. |n
    This function splits a dataset in the following way:|n
    |n
    |s|s1. Splits the dataset into 'train + val' and 'test' sets|n
    |s|s|s|s|sbased on person or object ID.|n
    |s|s2. Splits 'test' set into 'test-gallery' and 'test-query' sets|n
    |s|s|s|s|sin class-wise manner.|n
    |s|s3. Splits the 'train + val' set into 'train' and 'val' sets|n
    |s|s|s|s|sin the same way.|n
    |n
    The final subsets would be
    'train', 'val', 'test-gallery' and 'test-query'. |n
    |n
    Notes:|n
    |s|s- Each image is expected to have a single Label. Unlabeled or multi-labeled
    |s|s|s|simages will be split into 'not-supported'.|n
    |s|s- Object ID can be described by Label, or by attribute (--attr parameter)|n
    |s|s- The splits of the test set are controlled by '--query' parameter. |n
    |s|s|s|sGallery ratio would be 1.0 - query.|n
    |n
    Example: split a dataset in the specified ratio, split the test set|n
    into gallery and query in 1:1 ratio|n

    .. code-block::

    |s|s%(prog)s -t reidentification --subset train:.5 --subset val:.2 --subset test:.3 --query .5|n
    |n
    Example: use 'person_id' attribute for splitting|n

    .. code-block::

    |s|s%(prog)s --attr person_id
    """

    _default_query_ratio = 0.5

    def __init__(self, dataset, splits, query=None, attr_for_id=None, seed=None):
        """
        Parameters
        ----------
        dataset : Dataset
        splits : list
            A list of (subset(str), ratio(float))
            Subset is expected to be one of ["train", "val", "test"].
            The sum of ratios is expected to be 1.
        query : float
            The ratio of 'test-query' set.
            The ratio of 'test-gallery' set would be 1.0 - query.
        attr_for_id: str
            attribute name representing the person/object id.
            if this is not specified, label would be used.
        seed : int
            optional
        """
        super().__init__(dataset, splits, seed, restrict=True)

        if query is None:
            query = self._default_query_ratio

        assert 0.0 <= query and query <= 1.0, (
            "Query ratio is expected to be in the range " "[0, 1], but got %f" % query
        )
        test_splits = [("test-query", query), ("test-gallery", 1.0 - query)]

        # remove subset name restriction
        self._subsets = {"train", "val", "test-gallery", "test-query"}
        self._test_splits = test_splits
        self._attr_for_id = attr_for_id

    def _split_dataset(self):
        np.random.seed(self._seed)

        id_snames, id_ratio = self._snames, self._sratio

        attr_for_id = self._attr_for_id
        dataset = self._extractor

        # group by ID(attr_for_id)
        by_id = dict()
        annotations, unlabeled = self._get_uniq_annotations(dataset)
        if attr_for_id is None:  # use label
            for idx, ann in enumerate(annotations):
                ID = getattr(ann, "label", None)
                if ID not in by_id:
                    by_id[ID] = []
                by_id[ID].append((idx, ann))
        else:  # use attr_for_id
            for idx, ann in enumerate(annotations):
                attributes = dict(ann.attributes.items())
                assert attr_for_id in attributes, (
                    "'%s' is expected as an attribute name" % attr_for_id
                )
                ID = attributes[attr_for_id]
                if ID not in by_id:
                    by_id[ID] = []
                by_id[ID].append((idx, ann))

        required = self._get_required(id_ratio)
        if len(by_id) < required:
            log.warning(
                "There's not enough IDs, which is %s, "
                "so train/val/test ratio can't be guaranteed." % len(by_id)
            )

        # 1. split dataset into trval and test
        #    IDs in test set should not exist in train/val set.
        test = id_ratio[id_snames.index("test")] if "test" in id_snames else 0
        if NEAR_ZERO < test:  # has testset
            split_ratio = np.array([test, 1.0 - test])
            IDs = list(by_id.keys())
            np.random.shuffle(IDs)
            sections, _ = self._get_sections(len(IDs), split_ratio)
            splits = np.array_split(IDs, sections)
            testset = {pid: by_id[pid] for pid in splits[0]}
            trval = {pid: by_id[pid] for pid in splits[1]}
            # follow the ratio of datasetitems as possible.
            # naive heuristic: exchange the best item one by one.
            expected_count = int((len(self._extractor) - len(unlabeled)) * split_ratio[0])
            testset_total = int(np.sum([len(v) for v in testset.values()]))
            self._rebalancing(testset, trval, expected_count, testset_total)
        else:
            testset = dict()
            trval = by_id

        by_splits = dict()
        for subset in self._subsets:
            by_splits[subset] = []

        # 2. split 'test' into 'test-gallery' and 'test-query'
        if 0 < len(testset):
            test_snames = []
            test_ratio = []
            for sname, ratio in self._test_splits:
                test_snames.append(sname)
                test_ratio.append(float(ratio))

            self._split_by_attr(
                testset, test_snames, test_ratio, by_splits, merge_small_classes=False
            )

        # 3. split 'trval' into  'train' and 'val'
        trval_snames = ["train", "val"]
        trval_ratio = []
        for subset in trval_snames:
            if subset in id_snames:
                val = id_ratio[id_snames.index(subset)]
            else:
                val = 0.0
            trval_ratio.append(val)
        trval_ratio = np.array(trval_ratio)
        total_ratio = np.sum(trval_ratio)
        if total_ratio < NEAR_ZERO:
            trval_splits = list(zip(["train", "val"], trval_ratio))
            log.warning(
                "Sum of ratios is expected to be positive, "
                "got %s, which is %s" % (trval_splits, total_ratio)
            )
        else:
            trval_ratio /= total_ratio  # normalize
            self._split_by_attr(
                trval, trval_snames, trval_ratio, by_splits, merge_small_classes=False
            )

        # split unlabeled data into 'not-supported'.
        if len(unlabeled) > 0:
            self._subsets.add("not-supported")
            by_splits["not-supported"] = unlabeled

        self._set_parts(by_splits)

    @staticmethod
    def _rebalancing(test, trval, expected_count, testset_total):
        diffs = dict()
        for id_test, items_test in test.items():
            count_test = len(items_test)
            for id_trval, items_trval in trval.items():
                count_trval = len(items_trval)
                diff = count_trval - count_test
                if diff == 0:
                    continue  # exchange has no effect
                if diff not in diffs:
                    diffs[diff] = [(id_test, id_trval)]
                else:
                    diffs[diff].append((id_test, id_trval))
        if len(diffs) == 0:  # nothing would be changed by exchange
            return

        exchanges = []
        while True:
            target_diff = expected_count - testset_total
            # find nearest diff.
            keys = np.array(list(diffs.keys()))
            idx = (np.abs(keys - target_diff)).argmin()
            nearest = keys[idx]
            if abs(target_diff) <= abs(target_diff - nearest):
                break
            choice = np.random.choice(range(len(diffs[nearest])))
            id_test, id_trval = diffs[nearest][choice]
            testset_total += nearest
            new_diffs = dict()
            for diff, IDs in diffs.items():
                new_list = []
                for id1, id2 in IDs:
                    if id1 == id_test or id2 == id_trval:
                        continue
                    new_list.append((id1, id2))
                if 0 < len(new_list):
                    new_diffs[diff] = new_list
            diffs = new_diffs
            exchanges.append((id_test, id_trval))

        # exchange
        for id_test, id_trval in exchanges:
            test[id_trval] = trval.pop(id_trval)
            trval[id_test] = test.pop(id_test)


class _InstanceSpecificSplit(_TaskSpecificSplit):
    """
    Splits a dataset into subsets(train/val/test),
    using object annotations as a basis for splitting.|n
    Tries to produce an image split with the specified ratio, keeping the
    initial distribution of class objects.|n
    |n
    each image can have multiple object annotations -
    (instance bounding boxes, masks, polygons). Since an image shouldn't be included
    in multiple subsets at the same time, and image annotations
    shouldn't be split, in general, dataset annotations are unlikely to be split
    exactly in the specified ratio. |n
    This split tries to split dataset images as close as possible
    to the specified ratio, keeping the initial class distribution.|n
    |n
    Notes:|n
    |s|s- Each image is expected to have one or more annotations.|n
    |s|s- Only bbox annotations are considered in detection task.|n
    |s|s- Mask or Polygon annotations are considered in segmentation task.|n
    |n
    Example: split dataset so that each object class annotations were split|n
    in the specified ratio between subsets|n

    .. code-block::

    |s|s%(prog)s -t detection --subset train:.5 --subset val:.2 --subset test:.3 |n
    |s|s%(prog)s -t segmentation --subset train:.5 --subset val:.2 --subset test:.3
    """

    def __init__(self, dataset, splits, task, seed=None):
        """
        Parameters
        ----------
        dataset : Dataset
        splits : list
            A list of (subset(str), ratio(float))
            The sum of ratios is expected to be 1.
        seed : int
            optional
        """
        super().__init__(dataset, splits, seed)

        if task == SplitTask.detection.name:
            self.annotation_type = [AnnotationType.bbox]
        elif task == SplitTask.segmentation.name:
            self.annotation_type = [AnnotationType.mask, AnnotationType.polygon]

    def _group_by_labels(self, dataset):
        by_labels = dict()
        unlabeled = []

        for idx, item in enumerate(dataset):
            instance_anns = [a for a in item.annotations if a.type in self.annotation_type]
            if len(instance_anns) == 0:
                unlabeled.append(idx)
                continue
            for instance_ann in instance_anns:
                label = getattr(instance_ann, "label", None)
                if label not in by_labels:
                    by_labels[label] = [(idx, instance_ann)]
                else:
                    by_labels[label].append((idx, instance_ann))

        return by_labels, unlabeled

    def _split_dataset(self):
        np.random.seed(self._seed)

        subsets, sratio = self._snames, self._sratio

        # 1. group by bbox label
        by_labels, unlabeled = self._group_by_labels(self._extractor)

        # 2. group by attributes
        required = self._get_required(sratio)
        by_combinations = list()
        for _, items in by_labels.items():
            by_attributes = self._group_by_attr(items)
            # merge groups which have too small samples.
            attr_combinations = list(by_attributes.keys())
            np.random.shuffle(attr_combinations)  # add randomless
            cluster = []
            min_cluster = max(required, len(items) * 0.01)  # temp solution
            for attr in attr_combinations:
                indice = by_attributes[attr]
                if len(indice) >= min_cluster:
                    by_combinations.append(indice)
                else:
                    cluster.extend(indice)
                    if len(cluster) >= min_cluster:
                        by_combinations.append(cluster)
                        cluster = []

            if len(cluster) > 0:
                by_combinations.append(cluster)
                cluster = []

        total = len(self._extractor)
        # total number of GT samples per label-attr combinations
        n_combs = [len(v) for v in by_combinations]

        # 3-1. initially count per-image GT samples
        counts_all = {}
        for idx_img in range(total):
            if idx_img not in unlabeled:
                counts_all[idx_img] = dict()

        for idx_comb, indice in enumerate(by_combinations):
            for idx_img in indice:
                if idx_comb not in counts_all[idx_img]:
                    counts_all[idx_img][idx_comb] = 1
                else:
                    counts_all[idx_img][idx_comb] += 1

        by_splits = dict()
        for sname in self._subsets:
            by_splits[sname] = []

        target_ins = []  # target instance numbers to be split
        for sname, ratio in zip(subsets, sratio):
            target_ins.append([sname, np.array(n_combs) * ratio])

        init_scores = {}
        for idx_img, distributions in counts_all.items():
            norm_sum = 0.0
            for idx_comb, dis in distributions.items():
                norm_sum += dis / n_combs[idx_comb]
            init_scores[idx_img] = norm_sum

        by_scores = dict()
        for idx_img, score in init_scores.items():
            if score not in by_scores:
                by_scores[score] = [idx_img]
            else:
                by_scores[score].append(idx_img)

        # functions for keep the # of annotations not exceed the target_ins num
        def compute_penalty(counts, n_combs):
            p = 0
            for idx_comb, v in counts.items():
                if n_combs[idx_comb] <= 0:
                    p += 1
                else:
                    p += max(0, (v / n_combs[idx_comb]) - 1.0)

            return p

        def update_nc(counts, n_combs):
            for idx_comb, v in counts.items():
                n_combs[idx_comb] = n_combs[idx_comb] - v

        # 3-2. assign each DatasetItem to a split, one by one
        actual_ins = copy.deepcopy(target_ins)
        for score in sorted(by_scores.keys(), reverse=True):
            indice = by_scores[score]
            np.random.shuffle(indice)  # add randomness for the same score

            for idx in indice:
                counts = counts_all[idx]
                # shuffling split order to add randomness
                # when two or more splits have the same penalty value
                np.random.shuffle(actual_ins)

                pp = []
                for sname, nc in actual_ins:
                    if np.sum(nc) <= 0:
                        # the split has enough instances,
                        # stop adding more images to this split
                        pp.append(1e08)
                    else:
                        # compute penalty based on the number of GT samples
                        # added in the split
                        pp.append(compute_penalty(counts, nc))

                # we push an image to a split with the minimum penalty
                midx = np.argmin(pp)
                sname, nc = actual_ins[midx]
                by_splits[sname].append(idx)
                update_nc(counts, nc)

        # split unlabeled data
        if len(unlabeled) > 0:
            self._split_unlabeled(unlabeled, by_splits)

        self._set_parts(by_splits)



# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp
from shutil import rmtree

from datumaro.cli.util.errors import CliException
from datumaro.util.definitions import get_datumaro_cache_dir

from ..util import MultilineFormatter


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Generate synthetic dataset",
        description="""
        Creates a synthetic dataset with elements of the specified type and shape,
        and saves it in the provided directory.|n
        |n
        Currently, can only generate fractal images, useful for network compression.|n
        To create 3-channel images, you should provide the number of images, height and width.|n
        The images are colorized with a model, which will be downloaded automatically.|n
        Uses the algorithm from the article: https://arxiv.org/abs/2103.13023 |n
        |n
        Examples:|n
        - Generate 300 3-channel images with H=224, W=256 and store to data_dir:|n
        |s|s%(prog)s -o data_dir -k 300 --shape 224 256
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "-o", "--output-dir", required=True, help="Output directory to store generated dataset"
    )
    parser.add_argument(
        "-k", "--count", type=int, required=True, help="Number of images to be generated"
    )
    parser.add_argument(
        "--shape",
        nargs=2,
        metavar="DIM",
        type=int,
        required=True,
        help="Dimensions of data to be generated (height, width)",
    )
    parser.add_argument(
        "-t",
        "--type",
        default="image",
        choices=["image"],
        help="Specify type of data to generate (default: %(default)s)",
    )
    parser.add_argument(
        "--model-dir",
        type=str,
        default=get_datumaro_cache_dir(),
        help="Path to load the colorization model from. "
        "If no model is found, the model will be downloaded (default: %(default)s)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )

    parser.set_defaults(command=generate_command)

    return parser


def get_sensitive_args():
    return {generate_command: ["output_dir", "model_dir"]}


def generate_command(args):
    from datumaro.plugins.synthetic_data import FractalImageGenerator

    log.info("Generating dataset...")
    output_dir = args.output_dir

    if osp.isdir(output_dir) and os.listdir(output_dir):
        if args.overwrite:
            rmtree(output_dir)
            os.mkdir(output_dir)
        else:
            raise CliException(
                f"Directory '{output_dir}' already exists (pass --overwrite to overwrite)"
            )

    if args.type == "image":
        FractalImageGenerator(
            count=args.count, output_dir=output_dir, shape=args.shape, model_path=args.model_dir
        ).generate_dataset()
    else:
        raise NotImplementedError(f"Data type: {args.type} is not supported")

    log.info(f"Results have been saved to '{output_dir}'")

    return 0
